# pyflyby/test_autoimp.py

# License for THIS FILE ONLY: CC0 Public Domain Dedication
# http://creativecommons.org/publicdomain/zero/1.0/



import ast
import os
import pytest
from   shutil                   import rmtree
import sys
from   tempfile                 import mkdtemp
from   textwrap                 import dedent

from   pyflyby                  import (Filename, ImportDB, auto_eval,
                                        auto_import, find_missing_imports)
from   pyflyby._autoimp         import (LoadSymbolError, load_symbol,
                                        scan_for_import_issues)
from   pyflyby._flags           import CompilerFlags
from   pyflyby._idents          import DottedIdentifier
from   pyflyby._importstmt      import Import
from   pyflyby._util            import CwdCtx


@pytest.fixture
def tpp(request):
    """
    A temporary directory which is temporarily added to sys.path.
    """
    d = mkdtemp(prefix="pyflyby_test_autoimp_", suffix=".tmp")
    d = Filename(d).real
    def cleanup():
        # Unload temp modules.
        for name, module in sorted(sys.modules.items()):
            if (getattr(module, "__file__", None) or "").startswith(str(d)):
                del sys.modules[name]
        # Clean up sys.path.
        sys.path.remove(str(d))
        # Clean up directory on disk.
        rmtree(str(d))
    request.addfinalizer(cleanup)
    sys.path.append(str(d))
    return d


def writetext(filename, text, mode='w'):
    text = dedent(text)
    assert isinstance(filename, Filename)
    with open(str(filename), mode) as f:
        f.write(text)
    return filename


def _dilist2strlist(arg):
    assert type(arg) is list
    assert all(type(x) is DottedIdentifier for x in arg)
    return list(map(str, arg))


def test_find_missing_imports_basic_1():
    result   = find_missing_imports("os.path.join", namespaces=[{}])
    result   = _dilist2strlist(result)
    expected = ["os.path.join"]
    assert expected == result


def test_find_missing_imports_in_namespace_1():
    result   = find_missing_imports("os.path.join", namespaces=[{"os":os}])
    result   = _dilist2strlist(result)
    expected = []
    assert expected == result


def test_find_missing_imports_builtins_1():
    result   = find_missing_imports("os, sys, eval", [{"os": os}])
    result   = _dilist2strlist(result)
    expected = ['sys']
    assert expected == result


def test_find_missing_imports_undefined_1():
    result   = find_missing_imports("numpy.arange(x) + arange(y)", [{"y": 3}])
    result   = _dilist2strlist(result)
    expected = ['arange', 'numpy.arange', 'x']
    assert expected == result


def test_find_missing_imports_in_scope_1():
    result   = find_missing_imports("import numpy; numpy.arange(x) + arange(x)", [{}])
    result   = _dilist2strlist(result)
    expected = ['arange', 'x']
    assert expected == result


def test_find_missing_imports_in_scope_2():
    result   = find_missing_imports("from numpy import pi; numpy.pi + pi + x", [{}])
    result   = _dilist2strlist(result)
    expected = ['numpy.pi', 'x']
    assert expected == result


def test_find_missing_imports_in_scope_3():
    result   = find_missing_imports("for x in range(3): print(numpy.arange(x))", [{}])
    result   = _dilist2strlist(result)
    expected = ['numpy.arange']
    assert expected == result


def test_find_missing_imports_in_scope_funcall_1():
    result   = find_missing_imports("foo1 = func(); foo1.bar + foo2.bar", [{}])
    result   = _dilist2strlist(result)
    expected = ['foo2.bar', 'func']
    assert expected == result


def test_find_missing_imports_in_scope_assign_attr_1():
    result   = find_missing_imports("a.b.y = 1; a.b.x, a.b.y, a.b.z", [{}])
    result   = _dilist2strlist(result)
    expected = ['a.b.x', 'a.b.z']
    assert expected == result


def test_find_missing_imports_lambda_1():
    result   = find_missing_imports("(lambda x: x*x)(7)", [{}])
    result   = _dilist2strlist(result)
    expected = []
    assert expected == result


def test_find_missing_imports_lambda_2():
    result   = find_missing_imports("(lambda x: x*x)(7) + x", [{}])
    result   = _dilist2strlist(result)
    expected = ['x']
    assert expected == result


def test_find_missing_imports_lambda_3():
    result   = find_missing_imports("(lambda *a,**k: (a, k))(7, x=1)", [{}])
    result   = _dilist2strlist(result)
    expected = []
    assert expected == result


def test_find_missing_imports_list_comprehension_1():
    result   = find_missing_imports("[x+y+z for x,y in [(1,2)]], y", [{}])
    result   = _dilist2strlist(result)
    expected = ['y', 'z']
    assert expected == result


def test_find_missing_imports_list_comprehension_nested_tuple_1():
    result   = find_missing_imports("[w+x+y+z for x,(y,z) in []]", [{}])
    result   = _dilist2strlist(result)
    expected = ['w']
    assert expected == result


def test_find_missing_imports_list_comprehension_nested_tuple_2():
    result   = find_missing_imports(
        "[a+A+b+B+c+C+d+D+e+E+f+F+g+G for a,((b,c),d,[e,f,(g,)]) in []]", [{}])
    result   = _dilist2strlist(result)
    expected = ['A','B','C','D','E','F','G']
    assert expected == result


def test_find_missing_imports_generator_expression_1():
    result   = find_missing_imports("(x+y+z for x,y in [(1,2)]), y", [{}])
    result   = _dilist2strlist(result)
    expected = ['y', 'z']
    assert expected == result


def test_find_missing_imports_qualified_1():
    result   = find_missing_imports("( ( a . b ) . x ) . y + ( c + d ) . x . y", [{}])
    result   = _dilist2strlist(result)
    expected = ['a.b.x.y', 'c', 'd']
    assert expected == result


def test_find_missing_imports_ast_1():
    node = ast.parse("import numpy; numpy.arange(x) + arange(x)")
    result   = find_missing_imports(node, [{}])
    result   = _dilist2strlist(result)
    expected = ['arange', 'x']
    assert expected == result


def test_find_missing_imports_print_function_1():
    node = ast.parse(
        "from __future__ import print_function\n"
        "print (42, file=sys.stdout)\n"
    )
    result   = find_missing_imports(node, [{}])
    result   = _dilist2strlist(result)
    expected = ['sys.stdout']
    assert expected == result


def test_find_missing_imports_assignment_1():
    code = dedent("""
        from __future__ import print_function

        def f():
            x = 1
            print(x, y, z)
            y = 2
    """)
    result   = find_missing_imports(code, [{}])
    result   = _dilist2strlist(result)
    expected = ['y', 'z']
    assert expected == result


def test_find_missing_imports_function_body_1():
    code = dedent("""
        x1 = 1
        def func59399065():
            return x1 + x2 + x3
        x3 = 3
    """)
    result   = find_missing_imports(code, [{}])
    result   = _dilist2strlist(result)
    expected = ['x2']
    assert expected == result


def test_find_missing_imports_function_paramlist_1():
    code = dedent("""
        X1 = 1
        def func85025862(x1=X1, x2=X2, x3=X3):
            return x1 + x2 + x3
        X3 = 3
    """)
    result   = find_missing_imports(code, [{}])
    result   = _dilist2strlist(result)
    expected = ['X2', 'X3']
    assert expected == result


def test_find_missing_imports_function_paramlist_2():
    code = dedent("""
        from somewhere import given, data

        @given(data())
        def blargh(data):
            pass
    """)
    result  = find_missing_imports(code, [{}])
    result  = _dilist2strlist(result)
    expected = []
    assert expected == result

def test_find_missing_imports_function_defaults_1():
    code = dedent("""
        e = 1
        def func32466773(a=b, b=c, c=a, d=d, e=e, f=1):
            pass
    """)
    result   = find_missing_imports(code, [{}])
    result   = _dilist2strlist(result)
    expected = ['a', 'b', 'c', 'd']
    assert expected == result


def test_find_missing_imports_function_defaults_kwargs_1():
    code = dedent("""
        def func16049151(x=args, y=kwargs, z=y, *args, **kwargs):
            pass
    """)
    result   = find_missing_imports(code, [{}])
    result   = _dilist2strlist(result)
    expected = ['args', 'kwargs', 'y']
    assert expected == result


def test_find_missing_imports_kwarg_annotate():
    """
    pfb issue 162
    """
    code = dedent("""
        def func_pfb162(args:Dict, **kwargs:Any):
            pass
    """)
    result   = find_missing_imports(code, [{}])
    result   = _dilist2strlist(result)
    expected = ['Any', 'Dict']
    assert expected == result


def test_find_missing_imports_function_defaults_kwargs_2():
    code = dedent("""
        args = 1
        kwargs = 2
        def func69790319(x=args, y=kwargs, z=y, *args, **kwargs):
            pass
    """)
    result   = find_missing_imports(code, [{}])
    result   = _dilist2strlist(result)
    expected = ['y']
    assert expected == result


def test_find_missing_imports_function_paramlist_local_1():
    code = dedent("""
        x1 = 1
        x2 = 2
        def func77361554(x1, x3, x4):
            pass
        x4 = 4
        x1, x2, x3, x4, x5
    """)
    result   = find_missing_imports(code, [{}])
    result   = _dilist2strlist(result)
    expected = ['x3', 'x5']
    assert expected == result


def test_find_missing_imports_function_paramlist_selfref_1():
    code = dedent("""
        f1 = 'x'
        def f2(g1=f1, g2=f2, g3=f3):
            return (g1, g2, g3)
        f3 = 'x'
    """)
    result   = find_missing_imports(code, [{}])
    result   = _dilist2strlist(result)
    expected = ['f2', 'f3']
    assert expected == result


def test_find_missing_imports_function_paramlist_lambda_1():
    code = dedent("""
        X1 = 1
        def func85025862(x1=lambda: 1/X1, x2=lambda: 1/X2, x3=lambda: 1/X3):
            return x1() + x2() + x3()
        X3 = 3
    """)
    result   = find_missing_imports(code, [{}])
    result   = _dilist2strlist(result)
    expected = ['X2']
    assert expected == result


def test_find_missing_imports_decorator_1():
    code = dedent("""
        deco1 = 'x'
        @deco1
        @deco2
        @deco3
        def func33144383():
            pass
        deco3 = 'x'
    """)
    result   = find_missing_imports(code, [{}])
    result   = _dilist2strlist(result)
    expected = ['deco2', 'deco3']
    assert expected == result


def test_find_missing_imports_decorator_selfref_1():
    code = dedent("""
        deco = 'x'
        func1 = 'x'
        @deco(func1, func2, func3)
        def func2():
            pass
        func3 = 'x'
    """)
    result   = find_missing_imports(code, [{}])
    result   = _dilist2strlist(result)
    expected = ['func2', 'func3']
    assert expected == result


def test_find_missing_imports_decorator_paramlist_1():
    code = dedent("""
        p2 = 2
        def deco(*args): pass
        @deco(p1, p2, p3)
        def foo74632516():
            pass
        p3 = 3
    """)
    result   = find_missing_imports(code, [{}])
    result   = _dilist2strlist(result)
    expected = ['p1', 'p3']
    assert expected == result


def test_find_missing_imports_setattr_1():
    code = dedent("""
        aa = 1
        aa.xx.yy = 1
        bb.xx.yy = 1
    """)
    result   = find_missing_imports(code, [{}])
    result   = _dilist2strlist(result)
    # For now we intentionally don't auto-import 'bb' because that's more
    # likely to be a mistake.
    expected = []
    assert expected == result


def test_find_missing_imports_delattr_1():
    code = dedent("""
        foo1 = 1
        del foo1.bar, foo2.bar
    """)
    result   = find_missing_imports(code, [{}])
    result   = _dilist2strlist(result)
    expected = ['foo2']
    assert expected == result


def test_find_missing_imports_delitem_lhs_1():
    code = dedent("""
        foo1 = 1
        del foo1[123], foo2[123]
    """)
    result   = find_missing_imports(code, [{}])
    result   = _dilist2strlist(result)
    expected = ['foo2']
    assert expected == result


def test_find_missing_imports_delitem_rhs_1():
    code = dedent("""
        foo1 = 1
        bar1 = 1
        del foo1[bar1], foo1[bar2]
    """)
    result   = find_missing_imports(code, [{}])
    result   = _dilist2strlist(result)
    expected = ['bar2']
    assert expected == result

# TODO: unit tests for local, nonlocal


def test_find_missing_imports_classdef_1():
    code = dedent("""
        class Mahopac:
            pass
        class Gleneida(Mahopac):
            pass
        Mahopac, Carmel, Gleneida
    """)
    result   = find_missing_imports(code, [{}])
    result   = _dilist2strlist(result)
    expected = ['Carmel']
    assert expected == result


def test_find_missing_imports_class_base_1():
    code = dedent("""
        Mill = object
        class Mohansic(Crom, Mill):
            pass
    """)
    result   = find_missing_imports(code, [{}])
    result   = _dilist2strlist(result)
    expected = ['Crom']
    assert expected == result


def test_find_missing_imports_class_name_2():
    code = dedent(
        """
        class Decimal:
            def foo(self):

                Decimal.x = 1
    """
    )
    result = find_missing_imports(code, [{}])
    result = _dilist2strlist(result)
    expected = []
    assert expected == result


@pytest.mark.xfail(strict=True)
def test_find_missing_import_xfail_after_pr_152():
    code = dedent(
        """
        class MyClass(object):
            Outer = MyClass
    """
    )
    result = find_missing_imports(code, [{}])
    result = _dilist2strlist(result)
    expected = ["MyClass"]
    assert result == expected


def test_method_reference_current_class():
    """
    A method can reference the current class

    But only if this is a toplevel class, nesting won't work in Python.
    """
    code = dedent(
        """
       class Decimal:
           def foo(self):
               Decimal.x = 1
               Float.y=1

           class Real:

               def foo():
                   Real.r = 1
   """
    )
    missing, unused = scan_for_import_issues(code, [{}])
    # result = _dilist2strlist(result)
    assert missing == [
        (5, DottedIdentifier("Float.y")),
        (10, DottedIdentifier("Real.r")),
    ]
    assert unused == []


def test_annotation_inside_class():
    code = dedent(
        """
        class A:
            param1: str
            param2: B

        class B:
            param1: str
   """
    )
    missing, unused = scan_for_import_issues(code, [{}])
    assert missing == []
    assert unused == []


@pytest.mark.xfail(
    reason="Had to deactivate as part of https://github.com/deshaw/pyflyby/pull/269/files conflicting requirements",
    strict=True,
)
def test_find_missing_imports_class_name_1():
    code = dedent(
        """
        class Corinne:
            pass
        class Bobtail:
            class Chippewa:
                Bobtail # will be name error at runtime
            Rockton = Passall, Corinne, Chippewa
                      # ^error, ^ok   , ^ok
    """
    )
    result = find_missing_imports(code, [{}])
    result = _dilist2strlist(result)
    expected = ["Bobtail", "Passall"]
    assert expected == result


def test_find_missing_imports_class_name_1b():
    code = dedent(
        """
        class Corinne:
            pass
        class Bobtail:
            class Chippewa:
                Bobtail # will be name error at runtime
            Rockton = Passall, Corinne, Chippewa
                      # ^error, ^ok   , ^ok
    """
    )
    result = find_missing_imports(code, [{}])
    result = _dilist2strlist(result)
    expected = ["Passall"]
    assert expected == result



def test_find_missing_imports_class_members_1():
    code = dedent("""
        class Kenosha(object):
            x = 3
            z = x, y
    """)
    result   = find_missing_imports(code, [{}])
    result   = _dilist2strlist(result)
    expected = ['y']
    assert expected == result


def test_find_missing_imports_class_member_vs_function_1():
    code = dedent("""
        class Sidney(object):
            x = 3
            def barracuda(self):
                return x, y
    """)
    result   = find_missing_imports(code, [{}])
    result   = _dilist2strlist(result)
    expected = ['x', 'y']
    assert expected == result


def test_find_missing_imports_class_member_vs_function_2():
    code = dedent("""
        class Wayne: pass
        class Connaught(object):
            class Windsor: pass
            def Mercury(self):
                return Wayne, Connaught, Windsor
    """)
    result   = find_missing_imports(code, [{}])
    result   = _dilist2strlist(result)
    expected = ['Windsor']
    assert expected == result


def test_find_missing_imports_class_member_vs_lambda_1():
    code = dedent("""
        x1 = 1
        class Salsa(object):
            x2 = 2
            x3 = 3
            y = [lambda y3=x3, y4=x4: x1 + x2 + x5 + x6]
        x4 = 4
        x5 = 5
    """)
    result   = find_missing_imports(code, [{}])
    result   = _dilist2strlist(result)
    expected = ['x2', 'x4', 'x6']
    assert expected == result


def test_find_missing_imports_class_member_vs_paramlist_1():
    code = dedent("""
        class Drake:
            duck2 = 2
            def quack(self, mallard1=duck1, mallard2=duck2, mallard3=duck3):
                pass
            duck3 = 3
    """)
    result   = find_missing_imports(code, [{}])
    result   = _dilist2strlist(result)
    expected = ['duck1', 'duck3']
    assert expected == result


def test_find_missing_imports_class_member_vs_paramlist_lambda_1():
    code = dedent("""
        class Breakfast:
            def corn1(self):
                pass
            def cereal(self, maize1=lambda: corn1, maize2=lambda: corn2,
                             maize3=lambda: corn3):
                return (maize1(), maize2())
            def corn2(self):
                pass
        def corn3(self):
            pass
    """)
    result   = find_missing_imports(code, [{}])
    result   = _dilist2strlist(result)
    expected = ['corn1', 'corn2']
    assert expected == result


def test_find_missing_imports_class_member_vs_paramlist_local_1():
    code = dedent("""
        class Legume:
            x1 = 1
            x2 = 2
            def func13585710(x1, x3, x4):
                pass
            x4 = 4
            y = x1, x2, x3, x4, x5
    """)
    result   = find_missing_imports(code, [{}])
    result   = _dilist2strlist(result)
    expected = ['x3', 'x5']
    assert expected == result


def test_find_missing_imports_class_member_vs_decorator_1():
    code = dedent("""
        def deco(): 1/0
        class Cat:
            def panther1(self):
                pass
            @deco(panther1, panther2, panther3)
            def growl(self):
                pass
            def panther2(self):
                pass
    """)
    result   = find_missing_imports(code, [{}])
    result   = _dilist2strlist(result)
    expected = ['panther2', 'panther3']
    assert expected == result


def test_find_missing_imports_inner_class_method_1():
    code = dedent("""
        class Sand(object):
            Dirt = 100
            class Silicon:
                def f(self):
                    return Sand, Dirt, Silicon, Glass
        class Glass:
            pass
    """)
    result   = find_missing_imports(code, [{}])
    result   = _dilist2strlist(result)
    expected = ['Dirt', 'Silicon']
    assert expected == result


def test_find_missing_imports_inner_class_attribute_1():
    code = dedent("""
        x = 100
        class Axel(object):
            a = 100
            class Beth:
                b = x + a
    """)
    result   = find_missing_imports(code, [{}])
    result   = _dilist2strlist(result)
    expected = ['a']
    assert expected == result


def test_find_missing_imports_class_member_function_ref_1():
    code = dedent("""
        class Niska(object):
            def f1(self): pass
            g = f1, f2
    """)
    result   = find_missing_imports(code, [{}])
    result   = _dilist2strlist(result)
    expected = ['f2']
    assert expected == result


def test_find_missing_imports_class_member_generator_expression_1():
    # Verify that variables leak out of list comprehensions but not out of
    # generator expressions in Python 2.
    # Verify that both can see members of the same ClassDef.
    code = dedent("""
        class Caleb(object):
            x = []
            g1 = (1 for y1 in x)
            g2 = [1 for y2 in x]
            h = [y1, y2]
    """)
    result   = find_missing_imports(code, [{}])
    result   = _dilist2strlist(result)
    expected = ['y1', 'y2']
    assert expected == result


def test_find_missing_imports_latedef_def_1():
    code = dedent("""
        def marble(x):
            return x + y + z
        z = 100
    """)
    result   = find_missing_imports(code, [{}])
    result   = _dilist2strlist(result)
    expected = ['y']
    assert expected == result


def test_find_missing_imports_latedef_lambda_1():
    code = dedent("""
        granite = lambda x: x + y + z
        z = 100
    """)
    result   = find_missing_imports(code, [{}])
    result   = _dilist2strlist(result)
    expected = ['y']
    assert expected == result


def test_find_missing_imports_latedef_def_def_1():
    code = dedent("""
        def twodot():
            return sterling() + haymaker() + cannon() + twodot()
        def haymaker():
            return 100
    """)
    result   = find_missing_imports(code, [{}])
    result   = _dilist2strlist(result)
    expected = ['cannon', 'sterling']
    assert expected == result


def test_find_missing_imports_latedef_innerdef_1():
    code = dedent("""
        def kichawan(w):
            def turkey(x):
                return v + w + x + y + z
            z = 100
        v = 200
    """)
    result   = find_missing_imports(code, [{}])
    result   = _dilist2strlist(result)
    expected = ['y']
    assert expected == result


def test_find_missing_imports_latedef_innerdef_2():
    code = dedent("""
        def maple(w):
            def drumgor(x):
                return v + w + x + y + z
            z = 100
        def springmere(w):
            def dorchester(x):
                return v + w + x + y + z
        v = 200
    """)
    result   = find_missing_imports(code, [{}])
    result   = _dilist2strlist(result)
    expected = ['y', 'z']
    assert expected == result


def test_find_missing_imports_latedef_classdef_1():
    code = dedent("""
        a = 100
        class Granite:
            x = a, b
            def springs(self):
                x, y, z
        b = 100
        z = 100
    """)
    result   = find_missing_imports(code, [{}])
    result   = _dilist2strlist(result)
    expected = ['b', 'x', 'y']
    assert expected == result


def test_find_missing_imports_latedef_func_class_func_1():
    code = dedent("""
        def Nellie():
            class Shelley:
                def Norman(self):
                    return Alfred, Sherry, Grover, Kirk
            Sherry = 100
        Kirk = 200
    """)
    result   = find_missing_imports(code, [{}])
    result   = _dilist2strlist(result)
    expected = ['Alfred', 'Grover']
    assert expected == result


def test_find_missing_imports_latedef_if_1():
    code = dedent("""
        if 1:
            def cavalier():
                x, y
        if 1:
            x = 1
    """)
    result   = find_missing_imports(code, [{}])
    result   = _dilist2strlist(result)
    expected = ['y']
    assert expected == result


def test_find_missing_imports_class_scope_comprehension_1():
    code = dedent("""
        class Plymouth:
            x = []
            z = list(1 for t in x+y)
    """)
    result   = find_missing_imports(code, [{}])
    result   = _dilist2strlist(result)
    expected = ['y']
    assert expected == result


def test_find_missing_imports_global_1():
    code = dedent("""
        def func10663671():
            global x
            x = x + y
        x = 1
    """)
    result   = find_missing_imports(code, [{}])
    result   = _dilist2strlist(result)
    expected = ['y']
    assert expected == result


def test_find_missing_imports_complex_1():
    code = dedent("""
        x = 3+4j+5+k+u'a'
    """)
    result   = find_missing_imports(code, [{}])
    result   = _dilist2strlist(result)
    expected = ['k']
    assert expected == result


def test_find_missing_imports_code_1():
    f = lambda: foo.bar(x) + baz(y) # noqa: F821
    result   = find_missing_imports(f.__code__, [{}])
    result   = _dilist2strlist(result)
    expected = ['baz', 'foo.bar', 'x', 'y']
    assert expected == result


def test_find_missing_imports_code_args_1():
    def f(x, y, *a, **k):
        return g(x, y, z, a, k) # noqa: F821
    result   = find_missing_imports(f.__code__, [{}])
    result   = _dilist2strlist(result)
    expected = ['g', 'z']
    assert expected == result


def test_find_missing_imports_code_use_after_import_1():
    def f():
        import foo
        foo.bar()
    result   = find_missing_imports(f.__code__, [{}])
    result   = _dilist2strlist(result)
    expected = []
    assert expected == result


def test_find_missing_imports_code_lambda_scope_1():
    f = lambda x: (lambda: x+y) # noqa: F821
    result   = find_missing_imports(f.__code__, [{}])
    result   = _dilist2strlist(result)
    expected = ['y']
    assert expected == result


def test_find_missing_imports_code_conditional_1():
    def f():
        y0 = x0          # noqa: F821
        if c:            # noqa: F821
            y1 = y0 + x1 # noqa: F821
        else:
            y2 = x2 + y0 # noqa: F821
        x3 + y0          # noqa: F821
        y1 + y2
    result   = find_missing_imports(f.__code__, [{}])
    result   = _dilist2strlist(result)
    expected = ['c', 'x0', 'x1', 'x2', 'x3']
    assert expected == result


def test_find_missing_imports_code_loop_1():
    def f():
        for i in range(10):
            if i > 0:
                use(x)     # noqa: F821
                use(y)     # noqa: F821
            else:
                x = "hello" # noqa: F841
    result   = find_missing_imports(f.__code__, [{}])
    result   = _dilist2strlist(result)
    expected = ['use', 'y']
    assert expected == result


def test_find_missing_imports_positional_only_args_1():
    code = dedent("""
        def func(x, /, y):
            pass
    """)
    result   = find_missing_imports(code, [{}])
    result   = _dilist2strlist(result)
    expected = []
    assert expected == result


def test_find_missing_imports_keyword_only_args_1():
    code = dedent("""
        def func(*args, kwonly=b):
            a = kwonly
    """)
    result   = find_missing_imports(code, [{}])
    result   = _dilist2strlist(result)
    expected = ['b']
    assert expected == result


def test_find_missing_imports_keyword_only_args_2():
    code = dedent("""
        def func(*args, kwonly):
            a = kwonly
    """)
    result   = find_missing_imports(code, [{}])
    result   = _dilist2strlist(result)
    expected = []
    assert expected == result


def test_find_missing_imports_keyword_only_args_3():
    code = dedent("""
        def func(*args, kwonly, kwonly2=b):
            a = kwonly
    """)
    result   = find_missing_imports(code, [{}])
    result   = _dilist2strlist(result)
    expected = ['b']
    assert expected == result


def test_find_missing_imports_annotations_1():
    code = dedent("""
        def func(a: b) -> c:
            d = a
    """)
    result   = find_missing_imports(code, [{}])
    result   = _dilist2strlist(result)
    expected = ['b', 'c']
    assert expected == result


def test_find_missing_imports_annotations_2():
    code = dedent("""
    a: b = c
    """)
    result   = find_missing_imports(code, [{}])
    result   = _dilist2strlist(result)
    expected = ['b', 'c']
    assert expected == result


def test_find_missing_imports_star_assignments_1():
    code = dedent("""
    a, *b = c
    """)
    result   = find_missing_imports(code, [{}])
    result   = _dilist2strlist(result)
    expected = ['c']
    assert expected == result


def test_find_missing_imports_star_expression_1():
    code = dedent("""
    [a, *b]
    """)
    result   = find_missing_imports(code, [{}])
    result   = _dilist2strlist(result)
    expected = ['a', 'b']
    assert expected == result


def test_find_missing_imports_star_expression_2():
    code = dedent("""
    {a: 1, **b, **{c: 1}}
    """)
    result   = find_missing_imports(code, [{}])
    result   = _dilist2strlist(result)
    expected = ['a', 'b', 'c']
    assert expected == result


def test_find_missing_imports_star_expression_function_call_1():
    code = dedent("""
    f(a, *b, **c, d=e, **g)
    """)
    result   = find_missing_imports(code, [{}])
    result   = _dilist2strlist(result)
    expected = ['a', 'b', 'c', 'e', 'f', 'g']
    assert expected == result

def test_find_missing_imports_star_expression_function_call_2():
    code = dedent("""
    f(a, b=c, *d, **e)
    """)
    result   = find_missing_imports(code, [{}])
    result   = _dilist2strlist(result)
    expected = ['a', 'c', 'd', 'e', 'f']
    assert expected == result


def test_find_missing_imports_python_3_metaclass_1():
    code = dedent("""
    class Test(metaclass=TestMeta):
        pass
    """)
    result   = find_missing_imports(code, [{}])
    result   = _dilist2strlist(result)
    expected = ['TestMeta']
    assert expected == result


def test_find_missing_imports_f_string_1():
    code = dedent("""
    a = 1
    f'{a + 1} {b + 1}'
    """)
    result   = find_missing_imports(code, [{}])
    result   = _dilist2strlist(result)
    expected = ['b']
    assert expected == result


def test_find_missing_imports_f_string_2():
    code = dedent("""
    a = 1
    f'{a!s} {b!r}'
    """)
    result   = find_missing_imports(code, [{}])
    result   = _dilist2strlist(result)
    expected = ['b']
    assert expected == result


def test_find_missing_imports_f_string_3():
    # Recursive format spec
    code = dedent("""
    f'{a:{b}!s}'
    """)
    result   = find_missing_imports(code, [{}])
    result   = _dilist2strlist(result)
    expected = ['a', 'b']
    assert expected == result

def test_find_missing_imports_bytes_1():
    code = dedent("""
        a = b'b'
    """)
    result = find_missing_imports(code, [{}])
    result = _dilist2strlist(result)
    expected = []
    assert expected == result

def test_find_missing_imports_true_false_none_1():
    # These nodes changed in Python 3, make sure they are handled correctly
    code = dedent("""
    (True, False, None)
    """)
    result = find_missing_imports(code, [{}])
    result = _dilist2strlist(result)
    expected = []
    assert expected == result


@pytest.mark.skipif(sys.version_info < (3, 10), reason='No pattern matching before 3.10')
def test_find_missing_imports_pattern_match_1():
    code = dedent("""
    match {"foo": 1, "bar": 2}:
        case {
            "foo": the_foo_value,
            "bar": the_bar_value,
            **rest,
        }:
            print(the_foo_value)
        case _:
            pass
    """)
    result   = find_missing_imports(code, [{}])
    result   = _dilist2strlist(result)
    expected = []
    assert expected == result

@pytest.mark.xfail(reason='''The way the scope work in pyflyby it is hard to define a variable...
            only in one case I believe. We would need a scope stack in `def
            visit_match_case`, but that would remove the variable definition
            when leaving the match statement.
                   ''',strict=True)
@pytest.mark.skipif(sys.version_info < (3, 10), reason='No pattern matching before 3.10')
def test_find_missing_imports_pattern_match_2():
    code = dedent("""
    match {"foo": 1, "bar": 2}:
        case {
            "foo": the_foo_value,
            "bar": the_bar_value,
            **rest,
        }:
            print(the_foo_value)
        case _:
            print('here the_x_value might be unknown', the_foo_value)
    """)
    result   = find_missing_imports(code, [{}])
    result   = _dilist2strlist(result)
    expected = [DottedIdentifier('the_foo_value')]
    assert expected == result


def test_find_missing_imports_matmul_1():
    code = dedent("""
    a@b
    """)
    result   = find_missing_imports(code, [{}])
    result   = _dilist2strlist(result)
    expected = ['a', 'b']
    assert expected == result


def test_find_missing_imports_async_await_1():
    code = dedent("""
    async def f():
        async with a as b, c as d:
            async for i in e:
                g = await h()
    """)
    result   = find_missing_imports(code, [{}])
    result   = _dilist2strlist(result)
    expected = ['a', 'c', 'e', 'h']
    assert expected == result


def test_find_missing_imports_async_comprehension_1():
    code = dedent("""
    async def f():
        [i async for i in range(2)]
    """)
    result   = find_missing_imports(code, [{}])
    result   = _dilist2strlist(result)
    expected = []
    assert expected == result


def test_find_missing_imports_yield_from_1():
    code = dedent("""
    def f():
        yield from g()
    """)
    result   = find_missing_imports(code, [{}])
    result   = _dilist2strlist(result)
    expected = ['g']
    assert expected == result


def test_find_missing_imports_not_async_def():
    code = dedent(
        """
    async def f():
        pass
    f()
    """
    )
    result = find_missing_imports(code, [{}])
    expected = []
    assert expected == result


def test_find_missing_imports_nested_with_1():
    code = dedent("""
    with a as b, c as d:
        pass
    """)
    result   = find_missing_imports(code, [{}])
    result   = _dilist2strlist(result)
    expected = ['a', 'c']
    assert expected == result

def test_find_missing_imports_exception_1():
    code = dedent("""
    try:
        a = 1
    except:
        pass
    """)
    result = find_missing_imports(code, [{}])
    result = _dilist2strlist(result)
    expected = []
    assert expected == result

def test_find_missing_imports_exception_2():
    code = dedent("""
    try:
        a = 1
    except SomeException:
        pass
    """)
    result = find_missing_imports(code, [{}])
    result = _dilist2strlist(result)
    expected = ['SomeException']
    assert expected == result

def test_find_missing_imports_exception_3():
    code = dedent("""
    try:
        a = 1
    except SomeException as e:
        pass
    """)
    result = find_missing_imports(code, [{}])
    result = _dilist2strlist(result)
    expected = ['SomeException']
    assert expected == result


def test_find_missing_imports_tuple_ellipsis_type_1():
    code = dedent("""
    tuple[Foo, ..., Bar]
    """)
    result = find_missing_imports(code, [{}])
    result = _dilist2strlist(result)
    expected = ['Bar', 'Foo']
    assert expected == result


def test_scan_for_import_issues_type_comment_1():
    code = dedent("""
    from typing import Sequence
    def foo(strings  # type: Sequence[str]
            ):
        pass
    """)
    missing, unused = scan_for_import_issues(code)
    assert unused == []
    assert missing == []

ok1 = """
class MyClass:
    def get_class(self):
        return __class__
"""

ok2 = """
class MyClass:
    @classmethod
    def get_class(cls):
        return __class__
"""
ok3 = """
class MyClass:
    @staticmethod
    def get_class():
        return __class__
"""

@pytest.mark.parametrize('data', [ok1, ok2, ok3])
def test_fix_missing_dunder_class(data):
    """

    See https://github.com/deshaw/pyflyby/issues/325
    """
    code = dedent(data)
    missing = find_missing_imports(code, [{}])
    assert missing == []

notok1 = """print(__class__)"""
notok2 = """
class NotOk:
    print(__class__)"""


@pytest.mark.parametrize('data', [notok1, notok2])
def test_ok_dunder_class(data):
    """
    See https://github.com/deshaw/pyflyby/issues/325
    """
    code = dedent(data)
    missing = find_missing_imports(code, [{}])
    assert missing == [DottedIdentifier('__class__')]



def test_scan_for_import_issues_type_comment_2():
    code = dedent("""
    from typing import Sequence
    def foo(strings):
        # type: (Sequence[str]) -> None
        pass
    """)
    missing, unused = scan_for_import_issues(code)
    assert unused == []
    assert missing == []


def test_scan_for_import_issues_type_comment_3():
    code = dedent("""
    def foo(strings):
        # type: (Sequence[str]) -> None
        pass
    """)
    missing, unused = scan_for_import_issues(code)
    assert unused == []
    assert missing == [(1, DottedIdentifier('Sequence'))]


def test_scan_for_import_issues_type_comment_4():
    code = dedent("""
    from typing import Sequence, Tuple
    def foo(strings):
        # type: (Sequence[str]) -> None
        pass
    """)
    missing, unused = scan_for_import_issues(code)
    assert unused == [(2, Import('from typing import Tuple'))]
    assert missing == []


def test_scan_for_import_issues_multiline_string_1():
    code = dedent('''
    x = (
        """
        a
        """
        # blah
        "z"
    )
    ''')
    missing, unused = scan_for_import_issues(code)
    assert unused == []
    assert missing == []

def test_scan_for_import_issues_dictcomp_missing_1():
    code = dedent("""
        y1 = y2 = 1234
        {(x1,y1,z1): (x2,y2,z2) for x1,x2 in []}
    """)
    missing, unused = scan_for_import_issues(code)
    assert unused == []
    assert missing == [(3, DottedIdentifier('z1')), (3, DottedIdentifier('z2'))]


def test_scan_for_import_issues_dictcomp_unused_1():
    code = dedent("""
        import x1, x2, x3
        {123:x3 for x1,x2 in []}
    """)
    missing, unused = scan_for_import_issues(code)
    assert missing == []
    assert unused == [(2, Import('import x1')), (2, Import('import x2'))]


def test_scan_for_import_issues_setcomp_missing_1():
    code = dedent("""
        y1 = 1234
        {(x1,y1,z1) for x1,x2 in []}
    """)
    missing, unused = scan_for_import_issues(code)
    assert unused == []
    assert missing == [(3, DottedIdentifier('z1'))]


def test_scan_for_import_issues_setcomp_unused_1():
    code = dedent("""
        import x1, x2
        {x2 for x1 in []}
    """)
    missing, unused = scan_for_import_issues(code)
    assert missing == []
    assert unused == [(2, Import('import x1'))]


def test_scan_for_import_issues_class_subclass_imported_class_1():
    code = dedent("""
        from m1 import C1
        class C1(C1): pass
    """)
    missing, unused = scan_for_import_issues(code)
    assert missing == []
    assert unused == []


# This is now an XPASS, let's remove the xfail to see if it fails only on some CI version.
# # This is currently buggy.
# # The problem is that we postpone the check for C1.base = m1.C1, and by the
# # time we check it, we've already replaced the thing in the scope.
# @pytest.mark.xfail
def test_scan_for_import_issues_class_subclass_imported_class_in_func_1():
    code = dedent("""
        def f1():
            from m1 import C1
            class C1(C1): pass
    """)
    missing, unused = scan_for_import_issues(code)
    assert missing == []
    assert unused == []


def test_scan_for_import_issues_use_then_del_in_func_1():
    code = dedent("""
        def f1():
            x1 = 1
            x1
            del x1
    """)
    missing, unused = scan_for_import_issues(code)
    assert missing == []
    assert unused == []


def test_scan_for_import_issues_use_then_del_in_func_2():
    code = dedent("""
        def f1():
            x1 = 123
            print(x1); print(x2)
            del x1
    """)
    missing, unused = scan_for_import_issues(code)
    assert missing == [(4, DottedIdentifier('x2'))]
    assert unused == []


def test_scan_for_import_issues_del_in_func_then_use_1():
    code = dedent("""
        def f1():
            x1 = 123
            del x1
            print(x1); print(x2)
    """)
    missing, unused = scan_for_import_issues(code)
    assert missing == [(5, DottedIdentifier('x1')), (5, DottedIdentifier('x2'))]
    assert unused == []


def test_scan_for_import_issues_brace_identifiers_1():
    code = dedent("""
        import x1, x2, x3
        def f():
            '''{x1} {x3}'''
    """)
    missing, unused = scan_for_import_issues(code, parse_docstrings=True)
    assert missing == []
    assert unused == [(2, Import('import x2'))]


def test_scan_for_import_issues_brace_identifiers_bad_1():
    code = dedent("""
        import x1, x2, x3
        def f():
            '''{x1} {x3} {if}'''
    """)
    missing, unused = scan_for_import_issues(code, parse_docstrings=True)
    assert missing == []
    assert unused == [(2, Import('import x2'))]


def test_scan_for_import_issues_star_import_1():
    code = dedent("""
        import x1, y1
        x1, x2
        from m2 import *
        x3
    """)
    missing, unused = scan_for_import_issues(code, parse_docstrings=True)
    assert missing == [(3, DottedIdentifier('x2'))]
    assert unused == [(2, Import('import y1'))]


def test_scan_for_import_issues_star_import_deferred_1():
    code = dedent("""
        import x1, y1
        def f1():
            x1, x2
        from m2 import *
        def f1():
            x3
    """)
    missing, unused = scan_for_import_issues(code, parse_docstrings=True)
    assert missing == []
    assert unused == [(2, Import('import y1'))]


def test_scan_for_import_issues_star_import_local_1():
    code = dedent("""
        import x1, y1
        def f1():
            from m2 import *
            x2
        def f2():
            x3
        x1, x4
    """)
    missing, unused = scan_for_import_issues(code, parse_docstrings=True)
    assert missing == [(7, DottedIdentifier('x3')), (8, DottedIdentifier('x4'))]
    assert unused == [(2, Import('import y1'))]


def test_scan_for_import_issues_comprehension_subscript_1():
    code = dedent("""
        x = []
        [123 for x[0] in []]
    """)
    missing, unused = scan_for_import_issues(code, parse_docstrings=True)
    assert missing == []
    assert unused == []


def test_scan_for_import_issues_comprehension_subscript_missing_1():
    code = dedent("""
        [123 for x[0] in []]
    """)
    missing, unused = scan_for_import_issues(code, parse_docstrings=True)
    assert missing == [(2, DottedIdentifier('x'))]
    assert unused == []


def test_scan_for_import_issues_comprehension_subscript_complex_1():
    code = dedent("""
        dd = []
        [(aa,bb) for (bb, cc[0], dd[0]) in []]
    """)
    missing, unused = scan_for_import_issues(code, parse_docstrings=True)
    assert missing == [(3, DottedIdentifier('aa')),
                       (3, DottedIdentifier('cc'))]
    assert unused == []


def test_scan_for_import_issues_comprehension_attribute_1():
    code = dedent("""
        xx = []
        [123 for xx.yy in []]
    """)
    missing, unused = scan_for_import_issues(code, parse_docstrings=True)
    assert missing == []
    assert unused == []


@pytest.mark.xfail(strict=True)
def test_scan_for_import_issues_comprehension_attribute_missing_1():
    code = dedent("""
        [123 for xx.yy in []]
    """)
    missing, unused = scan_for_import_issues(code, parse_docstrings=True)
    assert missing == [(3, DottedIdentifier('xx'))]
    assert unused == []

# this is now an XPASS, check that it's passing everywhere in CI
# @pytest.mark.xfail
def test_scan_for_import_issues_comprehension_attribute_complex_1():
    code = dedent("""
        dd = []
        [(aa,bb) for (bb, cc.cx, dd.dx) in []]
    """)
    missing, unused = scan_for_import_issues(code, parse_docstrings=True)
    assert missing == [(3, DottedIdentifier('aa')),
                       (3, DottedIdentifier('cc.cx'))]
    assert unused == []


def test_scan_for_import_issues_dec_usage():
    """
    see https://github.com/deshaw/pyflyby/issues/265
    """
    code = dedent(
        """
        import random

        def repeat(times):
            def wrap(fn):
                return fn
            return wrap


        @repeat(random.choice([1, 4]))
        def foo(random):
            print(random)
    """
    )
    missing, unused = scan_for_import_issues(code, parse_docstrings=True)
    assert missing == []
    assert unused == []


def test_scan_for_import_issues_comprehension_attribute_subscript_1():
    code = dedent("""
        xx = []
        [123 for xx.yy[0] in []]
    """)
    missing, unused = scan_for_import_issues(code, parse_docstrings=True)
    assert missing == []
    assert unused == []


def test_scan_for_import_issues_comprehension_attribute_subscript_missing_1():
    code = dedent("""
        [123 for xx.yy[0] in []]
    """)
    missing, unused = scan_for_import_issues(code, parse_docstrings=True)
    assert missing == [(2, DottedIdentifier('xx.yy'))]
    assert unused == []


def test_scan_for_import_issues_comprehension_subscript_attribute_1():
    code = dedent("""
        xx = []
        [123 for xx[0].yy in []]
    """)
    missing, unused = scan_for_import_issues(code, parse_docstrings=True)
    assert missing == []
    assert unused == []


def test_scan_for_import_issues_comprehension_subscript_attribute_missing_1():
    code = dedent("""
        [123 for xx[0].yy in []]
    """)
    missing, unused = scan_for_import_issues(code, parse_docstrings=True)
    assert missing == [(2, DottedIdentifier('xx'))]
    assert unused == []


def test_scan_for_import_issues_generator_comprehension_subscript_attribute_1():
    code = dedent("""
        xx = []
        (123 for xx[0].yy in [])
    """)
    missing, unused = scan_for_import_issues(code, parse_docstrings=True)
    assert missing == []
    assert unused == []


def test_scan_for_import_issues_set_comprehension_subscript_attribute_1():
    code = dedent("""
        xx = []
        {123 for xx[0].yy in []}
    """)
    missing, unused = scan_for_import_issues(code, parse_docstrings=True)
    assert missing == []
    assert unused == []


def test_scan_for_import_issues_dict_comprehension_subscript_attribute_1():
    code = dedent("""
        xx = []
        {123:456 for xx[0].yy in []}
    """)
    missing, unused = scan_for_import_issues(code, parse_docstrings=True)
    assert missing == []
    assert unused == []


def test_scan_for_import_issues_setattr_1():
    code = dedent("""
        import aa, cc
        aa.xx.yy = 1
        bb.xx.yy = 1
    """)
    missing, unused = scan_for_import_issues(code)
    assert missing == [(4, DottedIdentifier('bb.xx.yy'))]
    # 'cc' should be marked as an unused-import, but 'aa' should be considered
    # used.  (This was buggy before 201907.)
    assert unused == [(2, Import('import cc'))]


def test_scan_for_import_issues_setattr_in_func_1():
    code = dedent("""
        import aa, cc
        def f():
            aa.xx.yy = 1
            bb.xx.yy = 1
    """)
    missing, unused = scan_for_import_issues(code)
    assert missing == [(5, DottedIdentifier('bb.xx.yy'))]
    # 'cc' should be marked as an unused-import, but 'aa' should be considered
    # used.  (This was buggy before 201907.)
    assert unused == [(2, Import('import cc'))]


def test_scan_for_import_issues_class_defined_after_use():
    code = dedent("""
    def foo():
        Bar.counter = 1

    class Bar:
        counter = 0
    """)
    missing, unused = scan_for_import_issues(code)
    assert missing == []
    assert unused == []


def test_setattr_is_not_unused():
    code = dedent("""
        from a import b
        def f():
            b.xx.yy = 1

    """)
    missing, unused = scan_for_import_issues(code)
    # b should be considered used
    assert missing == []
    assert unused == []


def test_all_exports_1():
    code = dedent("""
        from os import path, walk, read
        __all__ = ['path', 'rmdir', 'walk']
    """)
    missing, unused = scan_for_import_issues(code)
    # path and walk should not be unused
    assert missing == [(3, DottedIdentifier('rmdir'))]
    assert unused == [(2, Import('from os import read'))]


def test_all_exports_tuple_1():
    code = dedent("""
        from os import path, walk, read
        __all__ = ('path', 'rmdir', 'walk')
    """)
    missing, unused = scan_for_import_issues(code)
    assert missing == [(3, DottedIdentifier('rmdir'))]
    assert unused == [(2, Import('from os import read'))]


def test_all_exports_2():
    code = dedent("""
        from __future__ import absolute_import, division

        __all__ = ["defaultdict", "rmdir"]


        def defaultdict():
            pass
    """)
    missing, unused = scan_for_import_issues(code)
    assert missing == [(4, DottedIdentifier('rmdir'))]
    assert unused == []



def test_load_symbol_1():
    assert load_symbol("os.path.join", {"os": os}) is os.path.join


def test_load_symbol_2():
    assert load_symbol("os.path.join.__name__", {"os": os}) == "join"


def test_load_symbol_missing_1():
    with pytest.raises(LoadSymbolError):
        load_symbol("os.path.join.asdfasdf", {"os": os})


def test_load_symbol_missing_2():
    with pytest.raises(LoadSymbolError):
        load_symbol("os.path.join", {})


@pytest.mark.xfail(
    strict=True, reason="old test that was not names with test_ and never passed"
)
def test_load_symbol_eval_1():
    assert 'a/b' == load_symbol("os.path.join('a','b')", {"os": os})
    assert '/'   == load_symbol("os.path.join('a','b')[1]", {"os": os})
    assert 'A'   == load_symbol("os.path.join('a','b')[0].upper()", {"os": os})


@pytest.mark.xfail(
    strict=True, reason="old test that was not names with test_ and never passed"
)
def test_load_symbol_eval_2(capsys):
    assert '/' == load_symbol("(os.path.sep[0])", {}, autoimport=True,
                              allow_eval=True)
    out, _ = capsys.readouterr()
    expected = dedent("""
        [PYFLYBY] import os
    """).lstrip()
    assert expected == out


def test_load_symbol_no_eval_1():
    with pytest.raises(LoadSymbolError):
        load_symbol("os.path.join('a','b')", {"os": os})
    with pytest.raises(LoadSymbolError):
        load_symbol("os.path.join('a','b')[1]", {"os": os})
    with pytest.raises(LoadSymbolError):
        load_symbol("os.path.join('a','b')[0].upper", {"os": os})


def test_load_symbol_wrap_exc_1():
    class Foo89503828(object):
        def __getattr__(self, k):
            1/0
    ns = [{"foo": Foo89503828()}]
    try:
        load_symbol("foo.bar", ns)
    except LoadSymbolError as e:
        assert type(e.__cause__) == ZeroDivisionError
    else:
        assert False


@pytest.mark.xfail(
    strict=True, reason="old test that was not names with test_ and never passed"
)
def test_load_symbol_wrap_exc_eval_1():
    def foo31617859():
        return 1 / 0

    ns = [{"foo": foo31617859()}]
    try:
        load_symbol("foo()", ns, auto_eval=True)
    except LoadSymbolError as e:
        assert type(e.__cause__) == ZeroDivisionError
    else:
        assert False


@pytest.mark.xfail(
    strict=True, reason="old test that was not names with test_ and never passed"
)
def test_load_symbol_wrap_exc_eval_getattr_1():
    class Foo15356301(object):
        def __getattr__(self, k):
            1/0
    ns = [{"foo": Foo15356301()}]
    try:
        load_symbol("foo.bar", ns, auto_eval=True)
    except LoadSymbolError as e:
        assert type(e.__cause__) == ZeroDivisionError
    else:
        assert False


def test_auto_eval_1():
    result = auto_eval("b64decode('aGVsbG8=')")
    assert result == b'hello'


def test_auto_eval_locals_import_1():
    mylocals = {}
    result = auto_eval("b64decode('aGVsbG8=')", locals=mylocals)
    assert result == b'hello'
    assert mylocals["b64decode"] is __import__("base64").b64decode


def test_auto_eval_globals_import_1():
    myglobals = {}
    result = auto_eval("b64decode('aGVsbG8=')", globals=myglobals)
    assert result == b'hello'
    assert myglobals["b64decode"] is __import__("base64").b64decode


def test_auto_eval_custom_locals_1():
    result = auto_eval("b64decode('aGVsbG8=')",
                                   locals=dict(b64decode=lambda x: "blah"))
    assert result == 'blah'


def test_auto_eval_custom_globals_1():
    result = auto_eval("b64decode('aGVsbG8=')",
                                   globals=dict(b64decode=lambda x: "blah"))
    assert result == 'blah'


def test_auto_eval_exec_1():
    mylocals = dict(x=[])
    auto_eval("if True: x.append(b64decode('aGVsbG8='))",
              locals=mylocals)
    assert mylocals['x'] == [b'hello']
    assert mylocals["b64decode"] is __import__("base64").b64decode


def test_auto_eval_no_auto_flags_ps_flagps_1(capsys):
    auto_eval("print(3.00)", flags=CompilerFlags.from_int(0), auto_flags=False)
    out, _ = capsys.readouterr()
    assert out == "3.0\n"

def test_auto_eval_no_auto_flags_ps_flag_pf1():
    with pytest.raises(SyntaxError):
        auto_eval("print 3.00", flags="print_function", auto_flags=False)


def test_auto_eval_no_auto_flags_pf_flag_pf1(capsys):
    auto_eval("print(3.00, file=sys.stdout)",
              flags="print_function", auto_flags=False)
    out, _ = capsys.readouterr()
    assert out == "[PYFLYBY] import sys\n3.0\n"


def test_auto_eval_auto_flags_ps_flagps_1(capsys):
    with pytest.raises(SyntaxError):
        auto_eval("print 3.00", flags=CompilerFlags.from_int(0), auto_flags=True)


def test_auto_eval_auto_flags_pf_flagps_1(capsys):
    auto_eval("print(3.00, file=sys.stdout)", flags=CompilerFlags.from_int(0), auto_flags=True)
    out, _ = capsys.readouterr()
    assert out == "[PYFLYBY] import sys\n3.0\n"


def test_auto_eval_auto_flags_pf_flag_pf1(capsys):
    auto_eval("print(3.00, file=sys.stdout)",
              flags=CompilerFlags("print_function"), auto_flags=True)
    out, _ = capsys.readouterr()
    assert out == "[PYFLYBY] import sys\n3.0\n"


def test_auto_eval_proxy_module_1(tpp, capsys):
    os.mkdir("%s/tornado83183065"%tpp)
    writetext(tpp/"tornado83183065/__init__.py", """
        import sys
        twister = 54170888
        class P:
            def __getattr__(self, K):
                k = K.lower()
                if k == K:
                    raise AttributeError
                else:
                    return getattr(self, k)
        p = P()
        p.__dict__ = globals()
        p._m = sys.modules[__name__]
        sys.modules[__name__] = p
    """)
    writetext(tpp/"tornado83183065/hurricane.py", """
        cyclone = 79943637
    """)
    # Verify that we can auto-import a sub-module of a proxy module.
    result = auto_eval("tornado83183065.hurricane.cyclone")
    out, _ = capsys.readouterr()
    expected = dedent("""
        [PYFLYBY] import tornado83183065
        [PYFLYBY] import tornado83183065.hurricane
    """).lstrip()
    assert out == expected
    assert result == 79943637
    # Verify that the proxy module can do its magic stuff.
    result = auto_eval("tornado83183065.TWISTER")
    out, _ = capsys.readouterr()
    assert out == "[PYFLYBY] import tornado83183065\n"
    assert result == 54170888
    # Verify that the proxy module can do its magic stuff with a submodule
    # that's already imported.
    result = auto_eval("tornado83183065.HURRICANE.cyclone")
    out, _ = capsys.readouterr()
    assert out == "[PYFLYBY] import tornado83183065\n"
    assert result == 79943637


def test_auto_import_1(capsys):
    auto_import("sys.asdfasdf", [{}])
    out, _ = capsys.readouterr()
    expected = dedent("""
        [PYFLYBY] import sys
    """).lstrip()
    assert expected == out


def test_auto_import_multi_1(capsys):
    auto_import("sys.asdfasdf + os.asdfasdf", [{}])
    out, _ = capsys.readouterr()
    expected = dedent("""
        [PYFLYBY] import os
        [PYFLYBY] import sys
    """).lstrip()
    assert expected == out


def test_auto_import_nothing_1(capsys):
    auto_import("sys.asdfasdf", [{"sys":sys}])
    out, _ = capsys.readouterr()
    assert out == ""


def test_auto_import_some_1(capsys):
    auto_import("sys.asdfasdf + os.asdfasdf", [{"sys":sys}])
    out, _ = capsys.readouterr()
    expected = dedent("""
        [PYFLYBY] import os
    """).lstrip()
    assert expected == out


def test_auto_import_custom_1(tpp, capsys):
    writetext(tpp/"trampoline77069527.py", """
        print('hello  world')
    """)
    auto_import("trampoline77069527.asdfasdf", [{}])
    out, _ = capsys.readouterr()
    expected = dedent("""
        [PYFLYBY] import trampoline77069527
        hello  world
    """).lstrip()
    assert expected == out


def test_auto_import_custom_in_pkg_1(tpp, capsys):
    os.mkdir(str(tpp/"truck56331367"))
    writetext(tpp/"truck56331367/__init__.py", "")
    writetext(tpp/"truck56331367/tractor.py", """
        print('hello  there')
    """)
    auto_import("truck56331367.tractor", [{}])
    out, _ = capsys.readouterr()
    expected = dedent("""
        [PYFLYBY] import truck56331367
        [PYFLYBY] import truck56331367.tractor
        hello  there
    """).lstrip()
    assert expected == out


def test_auto_import_unknown_1(capsys):
    # Verify that if we try to access something that doesn't appear to be a
    # module, we don't attempt to import it (or at least don't log any visible
    # errors for it).
    auto_import("electron91631346.asdfasdf", [{}])
    out, _ = capsys.readouterr()
    assert out == ""


def test_auto_import_unknown_but_in_db1(tpp, capsys):
    # Verify that if we try to access something that's in the known-imports
    # database, but it doesn't actually exist, we get a visible error for it.
    db = ImportDB('import photon70447198')
    auto_import("photon70447198.asdfasdf", [{}], db=db)
    out, _ = capsys.readouterr()
    expected = dedent("""
        [PYFLYBY] import photon70447198
        [PYFLYBY] Error attempting to 'import photon70447198': ModuleNotFoundError: No module named 'photon70447198'
        Traceback (most recent call last):
    """).lstrip()

    assert out.startswith(expected)


def test_auto_import_fake_importerror_1(tpp, capsys):
    writetext(tpp/"proton24412521.py", """
        raise ImportError("No module named proton24412521")
    """)
    auto_import("proton24412521.asdfasdf", [{}])
    out, _ = capsys.readouterr()
    expected = dedent("""
        [PYFLYBY] import proton24412521
        [PYFLYBY] Error attempting to 'import proton24412521': ImportError: No module named proton24412521
        Traceback (most recent call last):
    """).lstrip()
    assert out.startswith(expected)


def test_auto_import_indirect_importerror_1(tpp, capsys):
    writetext(tpp/"neutron46291483.py", """
        import baryon96446873
    """)
    auto_import("neutron46291483.asdfasdf", [{}])
    out, _ = capsys.readouterr()
    expected = dedent("""
        [PYFLYBY] import neutron46291483
        [PYFLYBY] Error attempting to 'import neutron46291483': ModuleNotFoundError: No module named 'baryon96446873'
        Traceback (most recent call last):
    """).lstrip()

    assert out.startswith(expected)


def test_auto_import_nameerror_1(tpp, capsys):
    writetext(tpp/"lepton69688541.py", """
        foo
    """)
    auto_import("lepton69688541.asdfasdf", [{}])
    out, _ = capsys.readouterr()
    expected = dedent("""
        [PYFLYBY] import lepton69688541
        [PYFLYBY] Error attempting to 'import lepton69688541': NameError: name 'foo' is not defined
        Traceback (most recent call last):
    """).lstrip()
    assert out.startswith(expected)


def test_post_import_hook_incomplete_import():
    db = ImportDB('import glob')
    expected = Import('import glob')
    def test_hook(val):
        assert val == expected
    auto_import("glob.glob('*')", [{}], db=db, post_import_hook=test_hook)


def test_post_import_hook_alias_import():
    db = ImportDB('import numpy as np')
    expected = Import('import numpy as np')
    def test_hook(val):
        assert val == expected
    auto_import("np", [{}], db=db, post_import_hook=test_hook)


def test_post_import_hook_from_statement():
    db = ImportDB('from numpy import compat')
    expected = Import('from numpy import compat')
    def test_hook(val):
        assert val == expected
    auto_import("compat", [{}], db=db, post_import_hook=test_hook)


def test_post_import_hook_fullname():
    db = ImportDB('import numpy.compat')
    expected = Import('import numpy.compat')
    def test_hook(val):
        assert val == expected
    auto_import("numpy.compat", [{}], db=db, post_import_hook=test_hook)


def test_namespace_package(tpp, capsys):
    os.mkdir(str(tpp/'namespace_package'))
    auto_import("namespace_package", [{}])
    out, _ = capsys.readouterr()
    expected = dedent("""
        [PYFLYBY] import namespace_package
    """).lstrip()
    assert out.startswith(expected)


def test_unsafe_filename_warning(tpp, capsys):
    filepath = os.path.join(tpp._filename, 'foo#bar')
    os.mkdir(filepath)
    with CwdCtx(filepath):
        auto_import("pyflyby", [{}])
    out, _ = capsys.readouterr()
    expected = dedent("""
        [PYFLYBY] import pyflyby
    """).lstrip()
    assert out.startswith(expected)


def test_unsafe_filename_warning_II(tpp, capsys):
    filepath = os.path.join(tpp._filename, "foo#bar")
    os.mkdir(filepath)
    filepath = os.path.join(filepath, "qux#baz")
    os.mkdir(filepath)
    with CwdCtx(filepath):
        auto_import("pyflyby", [{}])
    out, _ = capsys.readouterr()
    expected = dedent(
        """
        [PYFLYBY] import pyflyby
    """
    ).lstrip()
    assert out.startswith(expected)



# pyflyby/test_modules.py

# License for THIS FILE ONLY: CC0 Public Domain Dedication
# http://creativecommons.org/publicdomain/zero/1.0/




import logging.handlers
from   pyflyby._file            import Filename
from   pyflyby._idents          import DottedIdentifier
from   pyflyby._modules         import ModuleHandle
import re
import subprocess
import sys
from   textwrap                 import dedent

import pytest

def test_ModuleHandle_1():
    m = ModuleHandle("sys")
    assert m.name == DottedIdentifier("sys")


def test_ModuleHandle_dotted_1():
    m = ModuleHandle("logging.handlers")
    assert m.name == DottedIdentifier("logging.handlers")


def test_ModuleHandle_from_module_1():
    m = ModuleHandle(logging.handlers)
    assert m == ModuleHandle("logging.handlers")
    assert m.name == DottedIdentifier("logging.handlers")


def test_eqne_1():
    m1a = ModuleHandle("foo.bar")
    m1b = ModuleHandle("foo.bar")
    m2  = ModuleHandle("foo.baz")
    assert     (m1a == m1b)
    assert not (m1a != m1b)
    assert not (m1a == m2)
    assert     (m1a != m2)


def test_filename_1():
    fn = logging.handlers.__file__
    fn = Filename(re.sub("[.]pyc$", ".py", fn)).real
    m = ModuleHandle("logging.handlers")
    assert m.filename.real == fn
    assert m.filename.base == "handlers.py"


def test_filename_init_1():
    fn = logging.__file__
    fn = Filename(re.sub("[.]pyc$", ".py", fn)).real
    m = ModuleHandle("logging")
    assert m.filename.real == fn
    assert m.filename.base == "__init__.py"


def test_module_1():
    m = ModuleHandle("logging")
    assert m.module is logging


# decimal used to be in there, but pytest + coverage seem to inject decimal
# in sys.modules
@pytest.mark.parametrize('modname', ['statistics', 'netrc'])
def test_filename_noload_1(modname):

    # PRE_TEST

    # Here we make sure that everything works properly before the actual test to
    # not get false positive


    # ensure there is no problem with sys.exit itself.
    ret = subprocess.run([sys.executable, '-c', dedent('''
        import sys
        sys.exit(0)
        ''')], capture_output=True)
    assert ret.returncode == 0, (ret, ret.stdout, ret.stderr)

    # Ensure there is no error with pyflyby itself
    ret = subprocess.run([sys.executable, '-c', dedent(f'''
        from pyflyby._modules import ModuleHandle
        import sys
        ModuleHandle("{modname}").filename
        sys.exit(0)
        ''')], capture_output=True)
    assert ret.returncode == 0, (ret, ret.stdout, ret.stderr)

    # ACTUAL TEST

    # don't exit with 1, as something else may exit with 1.
    ret = subprocess.run([sys.executable, '-c', dedent(f'''
        import sys
        if "{modname}" in sys.modules:
            sys.exit(120)
        from pyflyby._modules import ModuleHandle
        if "{modname}" in sys.modules:
            sys.exit(121)
        ModuleHandle("{modname}").filename
        if "{modname}" in sys.modules:
            sys.exit(123)
        else:
            sys.exit(0)
    ''')], capture_output=True)
    assert ret.returncode != 121, f"{modname} imported by pyflyby import"
    assert ret.returncode != 120, f"{modname} in sys.modules at startup"
    assert ret.returncode == 0, (ret, ret.stdout, ret.stderr)



# pyflyby/_idents.py.
# Copyright (C) 2011, 2012, 2013, 2014, 2018 Karl Chen.
# License: MIT http://opensource.org/licenses/MIT



from   functools                import total_ordering
from   keyword                  import kwlist
import re

from   pyflyby._util            import cached_attribute, cmp

from typing import Optional, Tuple, Dict


# Don't consider "print" a keyword, in order to be compatible with user code
# that uses "from __future__ import print_function".
_my_kwlist = list(kwlist)
_my_iskeyword = frozenset(_my_kwlist).__contains__


# TODO: use DottedIdentifier.prefixes
def dotted_prefixes(dotted_name, reverse=False):
    """
    Return the prefixes of a dotted name.

      >>> dotted_prefixes("aa.bb.cc")
      ['aa', 'aa.bb', 'aa.bb.cc']

      >>> dotted_prefixes("aa.bb.cc", reverse=True)
      ['aa.bb.cc', 'aa.bb', 'aa']

    :type dotted_name:
      ``str``
    :param reverse:
      If False (default), return shortest to longest.  If True, return longest
      to shortest.
    :rtype:
      ``list`` of ``str``
    """
    name_parts = dotted_name.split(".")
    if reverse:
        idxes = range(len(name_parts), 0, -1)
    else:
        idxes = range(1, len(name_parts)+1)
    result = ['.'.join(name_parts[:i]) or '.' for i in idxes]
    return result


def is_identifier(s: str, dotted: bool = False, prefix: bool = False):
    """
    Return whether ``s`` is a valid Python identifier name.

      >>> is_identifier("foo")
      True

      >>> is_identifier("foo+bar")
      False

      >>> is_identifier("from")
      False

    By default, we check whether ``s`` is a single valid identifier, meaning
    dots are not allowed.  If ``dotted=True``, then we check each dotted
    component::

      >>> is_identifier("foo.bar")
      False

      >>> is_identifier("foo.bar", dotted=True)
      True

      >>> is_identifier("foo..bar", dotted=True)
      False

      >>> is_identifier("foo.from", dotted=True)
      False

    By default, the string must comprise a valid identifier.  If
    ``prefix=True``, then allow strings that are prefixes of valid identifiers.
    Prefix=False excludes the empty string, strings with a trailing dot, and
    strings with a trailing keyword component, but prefix=True does not
    exclude these.

      >>> is_identifier("foo.bar.", dotted=True)
      False

      >>> is_identifier("foo.bar.", dotted=True, prefix=True)
      True

      >>> is_identifier("foo.or", dotted=True)
      False

      >>> is_identifier("foo.or", dotted=True, prefix=True)
      True

    :type s:
      ``str``
    :param dotted:
      If ``False`` (default), then the input must be a single name such as
      "foo".  If ``True``, then the input can be a single name or a dotted name
      such as "foo.bar.baz".
    :param prefix:
      If ``False`` (Default), then the input must be a valid identifier.  If
      ``True``, then the input can be a valid identifier or the prefix of a
      valid identifier.
    :rtype:
      ``bool``
    """
    if not isinstance(s, str):
        raise TypeError("is_identifier(): expected a string; got a %s"
                        % (type(s).__name__,))
    if prefix:
        return is_identifier(s + '_', dotted=dotted, prefix=False)
    if dotted:
        return all(is_identifier(w, dotted=False) for w in s.split('.'))
    return s.isidentifier() and not _my_iskeyword(s)


def brace_identifiers(text):
    """
    Parse a string and yield all tokens of the form "{some_token}".

      >>> list(brace_identifiers("{salutation}, {your_name}."))
      ['salutation', 'your_name']
    """
    if isinstance(text, bytes):
        text = text.decode('utf-8', errors='replace')
    for match in re.finditer("{([a-zA-Z_][a-zA-Z0-9_]*)}", text):
        yield match.group(1)


class BadDottedIdentifierError(ValueError):
    pass


# TODO: Use in various places, esp where e.g. dotted_prefixes is used.
@total_ordering
class DottedIdentifier:
    name: str
    parts: Tuple[str, ...]
    scope_info: Optional[Dict]

    def __new__(cls, arg, scope_info=None):
        if isinstance(arg, cls):
            return arg
        if isinstance(arg, str):
            return cls._from_name(arg, scope_info)
        if isinstance(arg, (tuple, list)):
            return cls._from_name(".".join(arg), scope_info)
        raise TypeError("DottedIdentifier: unexpected %s"
                        % (type(arg).__name__,))

    @classmethod
    def _from_name(cls, name, scope_info=None):
        self = object.__new__(cls)
        self.name = str(name)
        # TODO: change magic methods to compare with scopestack included
        self.scope_info = scope_info
        if not is_identifier(self.name, dotted=True):
            if len(self.name) > 20:
                raise BadDottedIdentifierError("Invalid python symbol name")
            else:
                raise BadDottedIdentifierError("Invalid python symbol name %r"
                                               % (name,))
        self.parts = tuple(self.name.split('.'))
        return self

    @cached_attribute
    def parent(self):
        if len(self.parts) > 1:
            return DottedIdentifier('.'.join(self.parts[:-1]))
        else:
            return None

    @cached_attribute
    def prefixes(self):
        parts = self.parts
        idxes = range(1, len(parts)+1)
        result = ['.'.join(parts[:i]) for i in idxes]
        return tuple(DottedIdentifier(x) for x in result)

    def startswith(self, o):
        o = type(self)(o)
        return self.parts[:len(o.parts)] == o.parts

    def __getitem__(self, x):
        return type(self)(self.parts[x])

    def __len__(self):
        return len(self.parts)

    def __iter__(self):
        return (type(self)(x) for x in self.parts)

    def __add__(self, suffix):
        return type(self)("%s.%s" % (self, suffix))

    def __str__(self):
        return self.name

    def __repr__(self):
        return "%s(%r)" % (type(self).__name__, self.name)

    def __hash__(self):
        return hash(self.name)

    def __eq__(self, other):
        if self is other:
            return True
        if not isinstance(other, DottedIdentifier):
            return NotImplemented
        return self.name == other.name

    def __ne__(self, other):
        if self is other:
            return False
        if not isinstance(other, DottedIdentifier):
            return NotImplemented
        return self.name != other.name

    # The rest are defined by total_ordering
    def __lt__(self, other):
        if not isinstance(other, DottedIdentifier):
            return NotImplemented
        return self.name < other.name

    def __cmp__(self, other):
        if self is other:
            return 0
        if not isinstance(other, DottedIdentifier):
            return NotImplemented
        return cmp(self.name, other.name)



# pyflyby/test_parse.py

# License for THIS FILE ONLY: CC0 Public Domain Dedication
# http://creativecommons.org/publicdomain/zero/1.0/



from   pyflyby._file            import FilePos, FileText, Filename
from   pyflyby._flags           import CompilerFlags
from   pyflyby._imports2s       import SourceToSourceFileImportsTransformation
from   pyflyby._parse           import PythonBlock, PythonStatement

import pytest
import sys
from   textwrap                 import dedent
import warnings

if sys.version_info < (3, 8, 3):
    print_function_flag = CompilerFlags.from_int(0x10000)
else:
    print_function_flag = CompilerFlags.from_int(0x100000)

def test_PythonBlock_FileText_1():
    text = FileText(
        dedent(
            """
        foo()
        bar()
    """
        ).lstrip(),
        filename="/foo/test_PythonBlock_1.py",
        startpos=(101, 55),
    )
    block = PythonBlock(text)
    assert block.annotated_ast_node
    assert text == block.text
    assert text == FileText(block)
    assert block.filename == Filename("/foo/test_PythonBlock_1.py")
    assert block.startpos == FilePos(101, 55)


def test_PythonBlock_firt_fstring():
    """
    Make sure fstring parsing work on Python 3.11 and before

    """
    block = PythonBlock(
        dedent(
            """
        some_fstring = f\"""abc
                    {', '.join(['''foo
                                   bar
                                   ''' for a in range(10)])}
\"""

    """
        ).strip()
    )
    assert block.annotated_ast_node


def test_PythonBlock_StatementComment():
    """
    Make sure that comments and statements in general do not start with a newline

    This would be problematic when reformatting files as we can't poke ahead at
    the next block to know whether or not it start a new line.
    """
    block = PythonBlock(dedent('''
        import foo

        # this

        import bar
    '''))
    for statement in block.statements:
        if statement.text.joined != '\n':
            assert not statement.text.joined.startswith('\n'), statement
    assert len(block.statements) == 5


def test_PythonBlock_attrs_1():
    block = PythonBlock(dedent('''
        foo()
        bar()
    ''').lstrip(), filename="/foo/test_PythonBlock_1.py", startpos=(101,99))
    assert block.annotated_ast_node
    assert block.text.lines == ("foo()", "bar()", "")
    assert block.filename   == Filename("/foo/test_PythonBlock_1.py")
    assert block.startpos   == FilePos(101, 99)


def test_PythonBlock_idempotent_1():
    block = PythonBlock("foo()\n")
    assert block.annotated_ast_node
    assert PythonBlock(block) is block


def test_PythonBlock_from_statement_1():
    stmt = PythonStatement("foo()\n", startpos=(101,202))
    block = PythonBlock(stmt)
    assert block.annotated_ast_node
    assert stmt.block is block
    assert block.statements == (stmt,)


def test_PythonBlock_statements_1():
    block = PythonBlock(dedent('''
        1
        [ 2,
          3,
        ]
        5
    ''').lstrip())
    expected = (
        PythonStatement("1\n"                            ),
        PythonStatement("[ 2,\n  3,\n]\n", startpos=(2,1)),
        PythonStatement("5\n"            , startpos=(5,1)),
    )
    assert block.statements == expected


def test_PythonBlock_lineno_1():
    block = PythonBlock(dedent('''
        1
        [ 2,
          3,
        ]
        5
    ''').lstrip(), startpos=(101,1))
    assert block.annotated_ast_node
    expected = (
        PythonStatement("1\n"            , startpos=(101,1)),
        PythonStatement("[ 2,\n  3,\n]\n", startpos=(102,1)),
        PythonStatement("5\n"            , startpos=(105,1)),
    )
    assert block.statements == expected


def test_PythonBlock_statements_comments_1():
    block = PythonBlock(dedent('''
        # 1
        print(2)
        # 3
        # 4
        print(5)
        x=[6,
         7]
        # 8
    ''').lstrip())
    expected = (
        PythonStatement('# 1\n'                 ),
        PythonStatement('print(2)\n',    startpos=(2,1)),
        PythonStatement('# 3\n# 4\n',   startpos=(3,1)),
        PythonStatement('print(5)\n',    startpos=(5,1)),
        PythonStatement('x=[6,\n 7]\n', startpos=(6,1)),
        PythonStatement('# 8\n',        startpos=(8,1)),
    )
    assert block.annotated_ast_node
    assert block.statements == expected


def test_PythonBlock_statements_continuation_1():
    block = PythonBlock(dedent(r'''
        a
        b,\
        c
        d
    ''').lstrip(), startpos=(101,1))
    expected = (
        PythonStatement('a\n' , startpos=(101,1)),
        PythonStatement('b,\\\nc\n', startpos=(102,1)),
        PythonStatement('d\n' , startpos=(104,1)),
    )
    assert block.annotated_ast_node
    assert block.statements == expected


def test_PythonBlock_statements_last_line_comment_continuation_1():
    block = PythonBlock(dedent(r'''
        a
        b\
        # c
    ''').lstrip(), startpos=(101,1))
    expected = (
        PythonStatement('a\n'       , startpos=(101,1)),
        PythonStatement('b\\\n# c\n', startpos=(102,1)),
    )
    assert block.annotated_ast_node
    assert block.statements == expected


def test_PythonBlock_statements_comment_no_continuation_1():
    block = PythonBlock(dedent('''
        x
        # y
        # z
    ''').lstrip())
    expected = (
        PythonStatement("x\n"                       ),
        PythonStatement("# y\n# z\n", startpos=(2,1)),
    )
    assert block.annotated_ast_node
    assert block.statements == expected


def test_PythonBlock_statements_comment_continuation_to_comment_1():
    block = PythonBlock(dedent('''
        x
        # y \\
        # z
    ''').lstrip())
    expected = (
        PythonStatement("x\n"                          ),
        PythonStatement("# y \\\n# z\n", startpos=(2,1)),
    )
    assert block.annotated_ast_node
    assert block.statements == expected


def test_PythonBlock_statements_last_line_no_newline_1():
    block = PythonBlock(dedent('''
        a
        b''').lstrip(), startpos=(101,1))
    expected = (
        PythonStatement('a\n', startpos=(101,1)),
        PythonStatement('b'  , startpos=(102,1)),
    )
    assert block.annotated_ast_node
    assert block.statements == expected


def test_PythonBlock_statements_last_line_comment_no_newline_1():
    block = PythonBlock(dedent('''
        a
        #b''').lstrip(), startpos=(101,1))
    expected = (
        PythonStatement('a\n', startpos=(101,1)),
        PythonStatement('#b'  , startpos=(102,1)),
    )
    assert block.annotated_ast_node
    assert block.statements == expected


def test_PythonBlock_statements_last_line_comment_continuation_no_newline_1():
    block = PythonBlock(dedent(r'''
        a
        b\
        # c''').lstrip(), startpos=(101,1))
    expected = (
        PythonStatement('a\n' , startpos=(101,1)),
        PythonStatement('b\\\n# c', startpos=(102,1)),
    )
    assert block.annotated_ast_node
    assert block.statements == expected


def test_PythonBlock_statements_last_line_nested_continuation_1():
    block = PythonBlock(dedent(r'''
        a
        if b:
            "c\
        # d"''').lstrip(), startpos=(101,1))
    expected = (
        PythonStatement('a\n' , startpos=(101,1)),
        PythonStatement('if b:\n    "c\\\n# d"', startpos=(102,1)),
    )
    assert block.annotated_ast_node
    assert block.statements == expected


def test_PythonBlock_statements_comment_backslash_1():
    block = PythonBlock(dedent(r'''
        #a\
        b''').lstrip(), startpos=(101,1))
    expected = (
        PythonStatement('#a\\\n', startpos=(101,1)),
        PythonStatement('b'     , startpos=(102,1)),
    )
    assert block.annotated_ast_node
    assert block.statements == expected


def test_PythonBlock_statements_single_1():
    block = PythonBlock("foo(1,2)")
    assert len(block.statements) == 1
    assert block.statements[0].block == block


def test_PythonBlock_statements_empty_1():
    block = PythonBlock("")
    assert len(block.statements) == 1
    assert block.statements[0].block == block


def test_PythonBlock_statements_lone_comment_no_newline_1():
    block = PythonBlock('#a a')
    assert len(block.statements) == 1
    assert block.statements[0].block == block


def test_PythonBlock_statements_lone_comment_no_newline_with_offset_1():
    block = PythonBlock('#a a', startpos=(101,55))
    assert len(block.statements) == 1
    assert block.statements[0].block == block


def test_PythonBlock_statements_single_trailing_comment_1():
    block = PythonBlock(dedent('''
        foo(1,2)
        # last
    ''').lstrip())
    expected = (
        PythonStatement("foo(1,2)\n", startpos=(1,1)),
        PythonStatement("# last\n"  , startpos=(2,1)),
    )
    assert block.statements == expected


def test_PythonBlock_statements_single_preceding_comment_1():
    block = PythonBlock(dedent('''
        # first
        foo(1,2)
    ''').lstrip())
    expected = (
        PythonStatement("# first\n" , startpos=(1,1)),
        PythonStatement("foo(1,2)\n", startpos=(2,1)),
    )
    assert block.statements == expected


def test_PythonBlock_statements_all_comments_1():
    block = PythonBlock("#a\n#b")
    assert len(block.statements) == 1
    assert block.statements[0].block == block


def test_PythonBlock_statements_all_comments_2():
    block = PythonBlock("#a\n#b\n")
    assert len(block.statements) == 1
    assert block.statements[0].block == block


def test_PythonBlock_doctest_1():
    block = PythonBlock(dedent("""
        # x
        '''
          >>> foo(bar
          ...     + baz)
        '''
    """).lstrip())
    expected = [PythonBlock('foo(bar\n    + baz)\n', startpos=(3,3))]
    assert block.get_doctests() == expected


def test_PythonBlock_doctest_nested_1():
    # Verify that we only include doctests from nested functions.
    block = PythonBlock(dedent("""
        def f():
            '>>> f(18739149)'
            def g():
                '>>> g(29355493)'
    """).lstrip())
    expected = [PythonBlock('f(18739149)\n', startpos=(2,5)),
                PythonBlock('g(29355493)\n', startpos=(4,9))]
    assert block.get_doctests() == expected


def test_PythonBlock_doctest_nested_cond_1():
    block = PythonBlock(dedent("""
        def f():
            '>>> f(17556901)'
            if True:
                def g():
                    '>>> g(21607865)'
    """).lstrip())
    expected = [PythonBlock('f(17556901)\n', startpos=(2,5)),
                PythonBlock('g(21607865)\n', startpos=(5,13))]
    assert block.get_doctests() == expected


def test_PythonBlock_doctest_nested_class_1():
    block = PythonBlock(dedent("""
        def f():
            '>>> f(11462083)'
            class C:
                '>>> C(21800340)'
                @classmethod
                def g(cls):
                    '>>> g(35606252)'
    """).lstrip())
    expected = [PythonBlock('f(11462083)\n', startpos=(2,5)),
                PythonBlock('C(21800340)\n', startpos=(4,9)),
                PythonBlock('g(35606252)\n', startpos=(7,13))]
    assert block.get_doctests() == expected


def test_PythonBlock_doctest_only_first_in_function_1():
    # Verify that we only include doctests from the first string in a
    # function.
    block = PythonBlock(dedent("""
        def f():
            '>>> a'
            3
            '>>> b'
    """).lstrip())
    expected = [PythonBlock('a\n', startpos=(2,5))]
    assert block.get_doctests() == expected


def test_PythonBlock_doctest_only_first_in_function_2():
    block = PythonBlock(dedent("""
        def f():
            if True:
                '>>> x'
    """).lstrip())
    expected = []
    assert block.get_doctests() == expected


def test_PythonBlock_doctest_only_first_in_function_3():
    block = PythonBlock(dedent("""
        def f():
            return '>>> x'
    """).lstrip())
    expected = []
    assert block.get_doctests() == expected


def test_PythonBlock_doctest_only_first_in_function_4():
    block = PythonBlock(dedent("""
        def f():
            ('>>> a' + '')
            3
    """).lstrip())
    expected = []
    assert block.get_doctests() == expected


def test_PythonBlock_doctest_only_first_in_function_not_try_1():
    block = PythonBlock(dedent("""
        def f():
            '>>> a'
            try:
                '>>> b'
            except:
                pass
    """).lstrip())
    expected = [PythonBlock('a\n', startpos=(2,5))]
    assert block.get_doctests() == expected


def test_PythonBlock_doctest_only_first_in_class_1():
    block = PythonBlock(dedent("""
        class C:
            '>>> C(11475111)'
            def f(self): pass
            '>>> x'

    """).lstrip())
    expected = [PythonBlock('C(11475111)\n', startpos=(2,5))]
    assert block.get_doctests() == expected


def test_PythonBlock_doctest_assignments_at_global_scope_1():
    # Verify that we only include doctests from (Epydoc) "variable docstrings"
    # at global scope.
    block = PythonBlock(dedent("""
        '>>> x'
        def f(): pass

        a = 4
        '>>> a'

        def g(): pass

        b = 5
        '>>> b'
    """).lstrip())
    expected = [PythonBlock('x\n'),
                PythonBlock('a\n', startpos=(5,1)),
                PythonBlock('b\n', startpos=(10,1))]
    assert block.get_doctests() == expected


def test_PythonBlock_doctest_assignments_ClassDef_1():
    block = PythonBlock(dedent("""
        class C:
            ">>> C(17621216)"
            x = 5
            ">>> C(28208124)"
            def f(self): pass
            ">>> x"
    """).lstrip())
    expected = [PythonBlock('C(17621216)\n', startpos=(2,5)),
                PythonBlock('C(28208124)\n', startpos=(4,5))]
    assert block.get_doctests() == expected


def test_PythonBlock_doctest_assignments_method_1():
    block = PythonBlock(dedent("""
        class C:
            ">>> C(13798505)"
            def __init__(self):
                ">>> C(25709748)"
                self.x = 0
                ">>> C(32231717)"
                f()
                ">>> x"
    """).lstrip())
    expected = [PythonBlock('C(13798505)\n', startpos=(2,5)),
                PythonBlock('C(25709748)\n', startpos=(4,9)),
                PythonBlock('C(32231717)\n', startpos=(6,9))]
    assert block.get_doctests() == expected


def test_PythonBlock_flags_good_1():
    PythonBlock('print("x",\n file=None)\n', flags="print_function").statements


def test_PythonBlock_flags_1():
    block = PythonBlock('print("x",\n file=None)\n', flags="print_function")
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", DeprecationWarning)
        assert block.flags == print_function_flag


def test_PythonBlock_flags_deduce_1():
    block = PythonBlock(
        dedent(
            """
        from __future__ import print_function
        print("x",
              file=None)
    """
        ).lstrip()
    )
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", DeprecationWarning)
        assert block.flags == print_function_flag


def test_PythonBlock_flags_type_comment_1():
    block = PythonBlock(
        dedent(
            """
    a = 1 # type: int
    b = None # type: ignore
    """
        ).lstrip()
    )
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", DeprecationWarning)
        assert block.flags == CompilerFlags(0x01000)


def test_PythonBlock_flags_type_comment_fail_transform():
    """
    See https://github.com/deshaw/pyflyby/issues/171

    
    $ python3.7 tidy-imports --print test.py
    def f():
        # type: () -> None
        pass

    $ python3.9 tidy-imports --print test.py
    Traceback (most recent call last):
    ....
    SNIP
    ....
      File ".../pyflyby/_parse.py", line 65, in _flatten_ast_nodes
        raise TypeError(
    TypeError: While processing /tmp/test.py: _flatten_ast_nodes: unexpected str

    /usr/local/bin/tidy-imports: encountered the following problems:
        /tmp/test.py: TypeError: _flatten_ast_nodes: unexpected str


    This is due to the fact that source-to-source does not support ast nodes being strings.
    """
    block = PythonBlock(
    dedent("""
     def f(x):
         # type:(str) -> str
         pass""")
    )

    s = SourceToSourceFileImportsTransformation(block)
    assert s.output() == block


examples_transform = [
    dedent(x)
    for x in [
        """
        a = None # type: ignore
        """,
        """
        class A:
            async def func(self, location: str) -> bytes:
                async with aiofiles.open(location, "rb") as file:
                    return await file.read()
        """,
        # positional only
        """
        def f(x, y=None, / , z=None):
             pass
        """,
    ]
]

if sys.version_info >= (3, 10):
    examples_transform.extend(
        [
            dedent(x)
            for x in [
                """
            match { "foo": 1, "bar": 2 }:
                case {
                    "foo": foo,
                    "bar": bar,
                    **rest,
                }:
                    pass
                case _:
                    pass
            """,
            """
            match event.get():
                case Click(position=(x, y)):
                    handle_click_at(x, y)
                case KeyPress(key_name="Q") | Quit():
                    game.quit()
                case KeyPress(key_name="up arrow"):
                    game.go_north()
                case KeyPress():
                    pass # Ignore other keystrokes
                case other_event:
                    raise ValueError(f"Unrecognized event: {other_event}")
            """,
            """
            match event.get():
                case Click((x, y), button=Button.LEFT):  # This is a left click
                    handle_click_at(x, y)
                case Click():
                    pass  # ignore other clicks
            """,
            """
            def http_error(status):
                match status:
                    case 400:
                        return "Bad request"
                    case 404:
                        return "Not found"
                    case 418:
                        return "I'm a teapot"
                    case 500 | 501 | 502:
                        return "I'm a teapot"
                    case _:
                        return "Something's wrong with the Internet"

            """,
                """
            match {"foo": 1, "bar": 2}:
                case {
                    "foo": foo,
                    "bar": bar,
                    **rest,
                }:
                    print(foo)
                case _:
                    pass
            """,
            ]
        ]
    )


@pytest.mark.parametrize("source", examples_transform)
def test_PythonBlock_flags_type_comment_ignore_fails_transform(source):
    """
    See https://github.com/deshaw/pyflyby/issues/174

    Type: ignore are custom ast.AST who have no col_offset.
    """
    block = PythonBlock(dedent(source))
    s = SourceToSourceFileImportsTransformation(block)
    assert s.output() == block



def test_PythonBlock_flags_deduce_eq_1():
    block1 = PythonBlock(
        dedent(
            """
        from __future__ import print_function
        print("x",
              file=None)
    """
        ).lstrip()
    )
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", DeprecationWarning)
        block2 = PythonBlock(
            dedent(
                """
            from __future__ import print_function
            print("x",
                  file=None)
        """
            ).lstrip(),
            flags=CompilerFlags.print_function,
        )
    assert block1 == block2


def test_PythonBlock_eqne_1():
    b1a = PythonBlock("foo()\nbar()\n")
    b1b = PythonBlock("foo()\nbar()\n")
    b2 = PythonBlock("foo()\nBAR()\n")
    assert b1a == b1b
    assert not (b1a != b1b)
    assert b1a != b2
    assert not (b1a == b2)


def test_PythonBlock_eqne_startpos_1():
    b1a = PythonBlock("foo()\nbar()\n")
    b1b = PythonBlock("foo()\nbar()\n", startpos=(1, 1))
    b2 = PythonBlock("foo()\nbar()\n", startpos=(1, 2))
    assert b1a == b1b
    assert not (b1a != b1b)
    assert b1a != b2
    assert not (b1a == b2)


def test_PythonBlock_eqne_flags_1():
    b1a = PythonBlock("foo", flags="print_function")
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", DeprecationWarning)
        b1b = PythonBlock("foo", flags=CompilerFlags.print_function)
    b2 = PythonBlock("foo")
    assert b1a == b1b
    assert not (b1a != b1b)
    assert b1a != b2
    assert not (b1a == b2)


def test_PythonStatement_from_source_1():
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", DeprecationWarning)
        stmt = PythonStatement(
            'print("x",\n file=None)\n'
        )
        assert stmt.block == PythonBlock(
            'print("x",\n file=None)\n'
        )


def test_PythonStatement_startpos_1():
    stmt = PythonStatement("foo()", startpos=(20, 30))
    assert stmt.startpos == FilePos(20, 30)
    assert stmt.block.startpos == FilePos(20, 30)
    assert stmt.block.text.startpos == FilePos(20, 30)


def test_PythonStatement_from_block_1():
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", DeprecationWarning)
        block = PythonBlock(
            'print("x",\n file=None)\n', flags=CompilerFlags.print_function
        )
    stmt = PythonStatement.from_block(block)
    assert stmt.block == block


def test_PythonStatement_bad_from_multi_statements_1():
    with pytest.raises(ValueError):
        PythonStatement("a\nb\n")


def test_PythonStatement_flags_good_1():
    PythonStatement('print("x",\n file=None)\n')


def test_str_lineno_simple_1():
    block = PythonBlock(dedent(r'''
        a = [
          2,
          3,
        ]
        b = 'five-a' + 'five-b'
        """
        seven
        eight
        """
        d = 'ten\n'
    ''').lstrip(), startpos=(101,1))
    expected_statements = (
        PythonStatement("a = [\n  2,\n  3,\n]\n"   , startpos=(101,1)),
        PythonStatement("b = 'five-a' + 'five-b'\n", startpos=(105,1)),
        PythonStatement('"""\nseven\neight\n"""\n' , startpos=(106,1)),
        PythonStatement("d = 'ten\\n'\n"           , startpos=(110,1))
    )
    assert block.statements == expected_statements
    literals = [(f.value, f.startpos) for f in block.string_literals()]
    expected = [
        ("five-a"          , FilePos(105,5)),
        ("five-b"          , FilePos(105,16)),
        ("\nseven\neight\n", FilePos(106,1)),
        ("ten\n"           , FilePos(110,5)),
    ]
    assert literals == expected


def test_str_lineno_trailing_1():
    block = PythonBlock(dedent(r'''
        one
        """
        three
        four
        """.split()
        six
    ''').lstrip(), startpos=(101,1))
    expected_statements = (
        PythonStatement('one\n'                          , startpos=(101,1)),
        PythonStatement('"""\nthree\nfour\n""".split()\n', startpos=(102,1)),
        PythonStatement('six\n'                          , startpos=(106,1)),
    )
    assert block.statements == expected_statements
    literals = [(f.value, f.startpos) for f in block.string_literals()]
    expected_literals = [("\nthree\nfour\n", FilePos(102,1))]
    assert literals == expected_literals


def test_str_lineno_multichars_1():
    block = PythonBlock(dedent('''
        [
          2,
          3,
        ]
        """ 5 "" " ""
        6 "" "
        7 "" """, """b "" "
        8 " "" """
        9
    ''').lstrip(), startpos=(101,1))
    expected_statements = (
        PythonStatement("[\n  2,\n  3,\n]\n", startpos=(101,1)),
        PythonStatement('""" 5 "" " ""\n6 "" "\n'
                        '7 "" """, """b "" "\n8 " "" """\n', startpos=(105,1)),
        PythonStatement("9\n", startpos=(109,1)),
        )
    assert block.statements == expected_statements
    literals = [(f.value, f.startpos) for f in block.string_literals()]
    expected_literals = [(' 5 "" " ""\n6 "" "\n7 "" ', FilePos(105,1 )),
                         ('b "" "\n8 " "" '          , FilePos(107,11))]
    assert literals == expected_literals


def test_str_lineno_in_dict_1():
    block = PythonBlock(dedent('''
        {
            1:
                """
                foo4
                bar5
                """,
            2:  \'\'\'
                foo8
                bar9
                \'\'\'}
    ''').lstrip(), startpos=(101,1))
    literals = [(f.value, f.startpos) for f in block.string_literals()]
    expected_literals = [('\n        foo4\n        bar5\n        ', FilePos(103,9)),
                         ('\n        foo8\n        bar9\n        ', FilePos(107,9))]
    assert literals == expected_literals


def test_str_lineno_strprefix_1():
    code = r'''
        r"aa\nbb"
        0
        r"x"
'''
    block = PythonBlock(dedent(code).lstrip(), startpos=(101,1))
    expected_statements = (
        PythonStatement('r"aa\\nbb"\n'       , startpos=(101,1)),
        PythonStatement('0\n'                , startpos=(102,1)),
        PythonStatement('r"x"\n'             , startpos=(103,1)),
    )
    assert block.statements == expected_statements
    literals = [(f.value, f.startpos) for f in block.string_literals()]
    expected_literals = [
        ('aa\\nbb'  , FilePos(101,1)),
        ('x'        , FilePos(103,1)),
    ]

    assert literals == expected_literals


def test_str_lineno_escaped_single_1():
    block = PythonBlock(dedent(r'''
        x = "a\
        #b" 'c\
        #d' \
        'e'
        'f' \
        #'g'
    ''').lstrip(), startpos=(101,1))
    expected_statements = (
        PythonStatement("""x = "a\\\n#b" 'c\\\n#d' \\\n'e'\n""", startpos=(101,1)),
        PythonStatement("""'f' \\\n#'g'\n""", startpos=(105,1)),
    )
    assert block.statements == expected_statements
    literals = [(f.value, f.startpos) for f in block.string_literals()]
    expected_literals = [('a#bc#de', FilePos(101,5)),
                         ('f'      , FilePos(105,1))]
    assert literals == expected_literals


def test_str_lineno_concatenated_1():
    code = '''
        "A" "a"
        "B" 'b'
        'C' 'c'
        'D' "d"
        """E
        e""" 'E'
        x = """F"""  'f'
        \'G\'\'\'\'g

        \'\'\' "G"
        x = """H
        h
        H""" 'h' \'\'\'H


        h\'\'\'
        "I" 'i'.split()
        "J" """j
        J"""    'j'.split()
        'K'
        'L'
        r"M" u'm'
        "N" ''"" "n" """"""\'\'\'\'\'\'\'N\'\'\'
        """
        O
        """
        """
        P
        """
        "Q" "q"""
        "R" "r""" + "S""""s""S""""s""""
        S"""
'''

    block = PythonBlock(dedent(code).lstrip(), startpos=(101,1))
    expected_statements = (
        PythonStatement('''"A" "a"\n'''            , startpos=(101,1)),
        PythonStatement('''"B" 'b'\n'''            , startpos=(102,1)),
        PythonStatement(''''C' 'c'\n'''            , startpos=(103,1)),
        PythonStatement(''''D' "d"\n'''            , startpos=(104,1)),
        PythonStatement('''"""E\ne""" 'E'\n'''     , startpos=(105,1)),
        PythonStatement('''x = """F"""  'f'\n'''   , startpos=(107,1)),
        PythonStatement("""'G''''g\n\n''' "G"\n""" , startpos=(108,1)),
        PythonStatement('''x = """H\nh\nH""" 'h' \'\'\'H\n\n\nh\'\'\'\n''', startpos=(111,1)),
        PythonStatement('''"I" 'i'.split()\n'''    , startpos=(117,1)),
        PythonStatement('''"J" """j\nJ"""    'j'.split()\n'''             , startpos=(118,1)),
        PythonStatement("""'K'\n"""                , startpos=(120,1)),
        PythonStatement("""'L'\n"""                , startpos=(121,1)),
        PythonStatement('''r"M" u\'m\'\n''' , startpos=(122,1)),
        PythonStatement('''"N" ''"" "n" """"""\'\'\'\'\'\'\'N\'\'\'\n'''  , startpos=(123,1)),
        PythonStatement('''"""\nO\n"""\n'''        , startpos=(124,1)),
        PythonStatement('''"""\nP\n"""\n'''        , startpos=(127,1)),
        PythonStatement('''"Q" "q"""\n'''          , startpos=(130,1)),
        PythonStatement('''"R" "r""" + "S""""s""S""""s""""\nS"""\n'''     , startpos=(131,1)),
    )

    assert block.statements == expected_statements
    literals = [(f.value, f.startpos) for f in block.string_literals()]
    expected_literals = [
        ("Aa", FilePos(101,1)),
        ("Bb", FilePos(102,1)),
        ("Cc", FilePos(103,1)),
        ("Dd", FilePos(104,1)),
        ("E\neE", FilePos(105,1)),
        ("Ff", FilePos(107,5)),
        ("Gg\n\nG", FilePos(108,1)),
        ("H\nh\nHhH\n\n\nh", FilePos(111,5)),
        ("Ii", FilePos(117,1)),
        ("Jj\nJj", FilePos(118,1)),
        ("K", FilePos(120,1)),
        ("L", FilePos(121,1)),
        ("Mm", FilePos(122,1)),
        ("NnN", FilePos(123,1)),
        ("\nO\n", FilePos(124,1)),
        ("\nP\n", FilePos(127,1)),
        ("Qq", FilePos(130,1)),
        ("Rr", FilePos(131,1)),
        ('Ss""Ss\nS', FilePos(131,13)),
    ]
    assert literals == expected_literals


def test_PythonBlock_compound_statements_1():
    block = PythonBlock(dedent('''
        foo(); bar()
        if condition:
            a; b
        {1:100,
         2:200};{3:300,
         4:400}; """
        foo
        bar
        """; """x
        y"""
    ''').lstrip(), startpos=(101,1))
    expected = (
        PythonStatement("foo(); "                  , startpos=(101,1)),
        PythonStatement("bar()\n"                  , startpos=(101,8)),
        PythonStatement("if condition:\n    a; b\n", startpos=(102,1)),
        PythonStatement("{1:100,\n 2:200};"        , startpos=(104,1)),
        PythonStatement("{3:300,\n 4:400}; "       , startpos=(105,9)),
        PythonStatement('"""\nfoo\nbar\n"""; '     , startpos=(106,10)),
        PythonStatement('"""x\ny"""\n'             , startpos=(109,6)),
    )
    assert block.statements == expected


def test_str_lineno_expression():
    # Code that used to be in test_interactive. _annotate_ast_startpos does
    # not work on it because it cannot handle multiline strings that contained
    # in a larger expression (they are joined implicitly because of the
    # parenthesis, even though the delimiters aren't on the same line). This
    # should start to work correctly in Python 3.8 because it marks the lineno
    # for multiline strings as the first line rather than the last. See issue #12.
    code = r'''
        ipython(
            # Verify that the auto importer isn't enabled yet.
            """
            In [1]: b64decode('x')
            ---------------------------------------------------------------------------
            NameError                                 Traceback (most recent call last)
            <ipython-input> in ...
            NameError: name 'b64decode' is not defined"""
            # Enable the auto importer.
            """
            In [2]: import pyflyby; pyflyby.enable_auto_importer()"""
            # Verify that the auto importer and tab completion work.
            """
            In [3]: b64deco\tde('aGF6ZWxudXQ=')
            [PYFLYBY] from base64 import b64decode
            Out[3]: 'hazelnut'
            """, args=['console'], kernel=kernel)
'''

    block = PythonBlock(dedent(code).lstrip())
    assert len(block.statements) == 1
    assert isinstance(block.statements[0], PythonStatement)


def test_PythonBlock_decorator_1():
    block = PythonBlock(dedent('''
        @foo1
        def bar1(): pass
        @foo2
        def bar2(): pass
    ''').lstrip(), startpos=(101,1))
    expected = (
        PythonStatement("@foo1\ndef bar1(): pass\n", startpos=(101,1)),
        PythonStatement("@foo2\ndef bar2(): pass\n", startpos=(103,1)),
    )
    assert block.statements == expected


def test_PythonBlock_with_1():
    block = PythonBlock(dedent('''
        with a: b
    ''').lstrip())
    expected = (PythonStatement("with a: b\n"),)
    assert block.statements == expected


def test_PythonBlock_with_2():
    block = PythonBlock(dedent('''
        with    closing(open('/etc/passwd')):
            pass
    ''').lstrip(), startpos=(101,1))
    expected = (
        PythonStatement("with    closing(open('/etc/passwd')):\n    pass\n",
                        startpos=(101,1)),
    )
    assert block.statements == expected


def test_PythonBlock_with_offset_1():
    block = PythonBlock(dedent('''
        with a:
            b
    ''').lstrip(), startpos=(101, 10))
    expected = (PythonStatement("with a:\n    b\n", startpos=(101, 10)),)
    assert block.statements == expected


def test_PythonBlock_doctest_with_1():
    block = PythonBlock(dedent('''
        def foo():
            """
            hello
              >>> with 11:
              ...   22
            """
    '''))
    doctest_blocks = block.get_doctests()
    doctest_block, = doctest_blocks
    expected = (PythonStatement("with 11:\n  22\n", startpos=(5, 11)),)
    assert doctest_block.statements == expected

def test_PythonBlock_doctest_ignore_doctest_options_1():
    block = PythonBlock(dedent('''
        def foo():
            """
            >>> 123 # doctest:+FOOBAR
            """
    '''))
    doctest_blocks = block.get_doctests()
    doctest_block, = doctest_blocks
    expected = (PythonStatement("123 # doctest:+FOOBAR\n", startpos=(4, 9)),)
    assert doctest_block.statements == expected


def test_PythonBlock_with_multi_1():
    block = PythonBlock(dedent('''
        with   A  as  a, B as b, C as c:
            pass
    ''').lstrip(), startpos=(101,1))
    expected = (
        PythonStatement("with   A  as  a, B as b, C as c:\n    pass\n",
                        startpos=(101,1)),
    )
    assert block.statements == expected


# auto_flags tests.
# Key for test names:
#   ps = code contains print statement
#   pf = code contains print function
#   pn = code does not contain print statement or function
#   px = code contains ambiguous print statement/function
#   flagps = flags does not contain CompilerFlags("print_function")
#   flagpf = flags contains CompilerFlags("print_function")
#   futpf = code contains 'from __future__ import print_function'


def test_PythonBlock_no_auto_flags_ps_flagpf_1():
    block = PythonBlock(dedent('''
        print 42
    ''').lstrip(), flags="print_function")
    with pytest.raises(SyntaxError):
        block.ast_node


def test_PythonBlock_no_auto_flags_pf_flagpf_1():
    block = PythonBlock(dedent('''
        print(42, out=x)
    ''').lstrip(), flags="print_function")
    assert     (block.flags                & "print_function")
    assert     (block.ast_node.input_flags & "print_function")
    assert not (block.source_flags         & "print_function")


def test_PythonBlock_no_auto_flags_pf_flagps_futpf_1():
    block = PythonBlock(dedent('''
        from __future__ import print_function
        print(42)
    ''').lstrip())
    assert     (block.flags                & "print_function")
    assert not (block.ast_node.input_flags & "print_function")
    assert     (block.source_flags         & "print_function")


def test_PythonBlock_no_auto_flags_pf_flagpf_futpf_1():
    block = PythonBlock(dedent('''
        from __future__ import print_function
        print(42)
    ''').lstrip(), flags="print_function")
    assert     (block.flags                & "print_function")
    assert     (block.ast_node.input_flags & "print_function")
    assert     (block.source_flags         & "print_function")


def test_PythonBlock_no_auto_flags_px_flagps_1():
    block = PythonBlock(dedent('''
        print(42)
    ''').lstrip())
    assert not (block.flags                & "print_function")
    assert not (block.ast_node.input_flags & "print_function")
    assert not (block.source_flags         & "print_function")


def test_PythonBlock_no_auto_flags_px_flagpf_1():
    block = PythonBlock(dedent('''
        print(42)
    ''').lstrip(), flags="print_function")
    assert     (block.flags                & "print_function")
    assert     (block.ast_node.input_flags & "print_function")
    assert not (block.source_flags         & "print_function")


def test_PythonBlock_no_auto_flags_pn_flagps_1():
    block = PythonBlock(dedent('''
        42
    ''').lstrip())
    assert not (block.flags                & "print_function")
    assert not (block.ast_node.input_flags & "print_function")
    assert not (block.source_flags         & "print_function")


def test_PythonBlock_no_auto_flags_pn_flagpf_1():
    block = PythonBlock(dedent('''
        42
    ''').lstrip(), flags="print_function")
    assert     (block.flags                & "print_function")
    assert     (block.ast_node.input_flags & "print_function")
    assert not (block.source_flags         & "print_function")


def test_PythonBlock_no_auto_flags_pn_futpf_1():
    block = PythonBlock(dedent('''
        from __future__ import print_function
        42
    ''').lstrip())
    assert     (block.flags                & "print_function")
    assert not (block.ast_node.input_flags & "print_function")
    assert     (block.source_flags         & "print_function")


def test_PythonBlock_auto_flags_pf_flagps_1():
    block = PythonBlock(dedent('''
        print(42, out=x)
    ''').lstrip(), auto_flags=True)
    assert not (block.flags                & "print_function")
    assert not (block.ast_node.input_flags & "print_function")
    assert not (block.source_flags         & "print_function")


def test_PythonBlock_auto_flags_pf_flagpf_1():
    block = PythonBlock(dedent('''
        print(42, out=x)
    ''').lstrip(), flags="print_function", auto_flags=True)
    assert     (block.flags                & "print_function")
    assert     (block.ast_node.input_flags & "print_function")
    assert not (block.source_flags         & "print_function")


def test_PythonBlock_auto_flags_pf_flagps_futpf_1():
    block = PythonBlock(dedent('''
        from __future__ import print_function
        print(42, out=x)
    ''').lstrip(), auto_flags=True)
    assert     (block.flags                & "print_function")
    assert not (block.ast_node.input_flags & "print_function")
    assert     (block.source_flags         & "print_function")


def test_PythonBlock_auto_flags_pf_flagpf_futpf_1():
    block = PythonBlock(dedent('''
        from __future__ import print_function
        print(42, out=x)
    ''').lstrip(), flags="print_function", auto_flags=True)
    assert     (block.flags                & "print_function")
    assert     (block.ast_node.input_flags & "print_function")
    assert     (block.source_flags         & "print_function")


def test_PythonBlock_auto_flags_px_flagps_1():
    block = PythonBlock(dedent('''
        print(42)
    ''').lstrip(), auto_flags=True)
    assert not (block.flags                & "print_function")
    assert not (block.ast_node.input_flags & "print_function")
    assert not (block.source_flags         & "print_function")


def test_PythonBlock_auto_flags_px_flagpf_1():
    block = PythonBlock(dedent('''
        print(42)
    ''').lstrip(), flags="print_function", auto_flags=True)
    assert     (block.flags                & "print_function")
    assert     (block.ast_node.input_flags & "print_function")
    assert not (block.source_flags         & "print_function")


def test_PythonBlock_auto_flags_px_flagps_futpf_1():
    block = PythonBlock(dedent('''
        from __future__ import print_function
        print(42)
    ''').lstrip(), auto_flags=True)
    assert     (block.flags                & "print_function")
    assert not (block.ast_node.input_flags & "print_function")
    assert     (block.source_flags         & "print_function")


def test_PythonBlock_auto_flags_px_flagpf_futpf_1():
    block = PythonBlock(dedent('''
        from __future__ import print_function
        print(42)
    ''').lstrip(), flags="print_function", auto_flags=True)
    assert     (block.flags                & "print_function")
    assert     (block.ast_node.input_flags & "print_function")
    assert     (block.source_flags         & "print_function")


def test_PythonBlock_auto_flags_pn_flagps_1():
    block = PythonBlock(dedent('''
        42
    ''').lstrip(), auto_flags=True)
    assert not (block.flags                & "print_function")
    assert not (block.ast_node.input_flags & "print_function")
    assert not (block.source_flags         & "print_function")


def test_PythonBlock_auto_flags_pn_flagpf_1():
    block = PythonBlock(dedent('''
        42
    ''').lstrip(), flags="print_function", auto_flags=True)
    assert     (block.flags                & "print_function")
    assert     (block.ast_node.input_flags & "print_function")
    assert not (block.source_flags         & "print_function")


def test_PythonBlock_auto_flags_pn_flagps_futpf_1():
    block = PythonBlock(dedent('''
        from __future__ import print_function
        42
    ''').lstrip(), auto_flags=True)
    assert     (block.flags                & "print_function")
    assert not (block.ast_node.input_flags & "print_function")
    assert     (block.source_flags         & "print_function")


def test_PythonBlock_auto_flags_pn_flagpf_futpf_1():
    block = PythonBlock(dedent('''
        from __future__ import print_function
        42
    ''').lstrip(), flags="print_function", auto_flags=True)
    assert     (block.flags                & "print_function")
    assert     (block.ast_node.input_flags & "print_function")
    assert     (block.source_flags         & "print_function")


def test_PythonStatement_flags_1():
    block = PythonBlock("from __future__ import unicode_literals\nx\n",
                        flags="division")
    s0, s1 = block.statements
    assert s0.block.source_flags == CompilerFlags("unicode_literals")
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", DeprecationWarning)
        assert s1.block.source_flags == CompilerFlags(0)
    assert s0.block.flags == CompilerFlags("unicode_literals", "division",)
    assert s1.block.flags == CompilerFlags("unicode_literals", "division",)


def test_PythonStatement_auto_flags_1():
    block = PythonBlock(
        "from __future__ import unicode_literals\nprint(1,file=x)\n",
        flags="division", auto_flags=True)
    s0, s1 = block.statements
    assert s0.block.source_flags == CompilerFlags("unicode_literals")
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", DeprecationWarning)
        assert s1.block.source_flags == CompilerFlags(0)
    expected = CompilerFlags("unicode_literals", "division")
    assert s0.block.flags        == expected
    assert s1.block.flags        == expected


def test_parsable_1():
    block = PythonBlock("if 1:\n  2")
    assert block.parsable


def test_unparsable_1():
    block = PythonBlock("if 1:\n2")
    assert not block.parsable


def test_parsable_explicit_flags_1():
    block = PythonBlock("print(3, file=4)", flags="print_function")
    assert block.parsable


def test_parsable_missing_flags_auto_flags_1():
    block = PythonBlock("print(3, file=4)", auto_flags=True)
    assert block.parsable


@pytest.mark.parametrize(
    "input",
    [
        "print(abc=123, *args, **kwargs)",
        "print(*args, ijk=123, **kwargs)",
        "print(7, *args, **kwargs)",
        "print(*args, 12, **kwargs)",
    ],
)
def test_parsable_Call_Ast_args_kwargs(input):
    block = PythonBlock(input, auto_flags=True)
    assert block.annotated_ast_node


@pytest.mark.parametrize(
    "input",
    [
        """
def func(x: List[int], y: List[int]) -> List[int]:
    return x + y
"""
    ],
)
def test_parsable_annotation_order(input):
    block = PythonBlock(input, auto_flags=True)
    assert block.annotated_ast_node


@pytest.mark.parametrize(
    "input",
    [
        '''x='abc'
f"""\
This is a
multi-line
f-string
with input {x}.\
"""
''',
        '''x=123
f"""\
{x}
is the value x
"""
''',
        '''x=456
f"""{x}
is the value x
"""
''',
        '''x=123
f"""\
In the middle
{x}
is the value.
"""
''',
    ],
)
def test_parse_f_string_ast_ann(input):
    block = PythonBlock(input, auto_flags=True)
    assert block.annotated_ast_node


@pytest.mark.parametrize(
    "input",
    [
        """
x = 123
fail_here = f"{x.stem} is no-op. \\
(Do we need to delete this still?)"
""",'''
BLAH = "blah"

def f(t):
    pass

f(f"""
{BLAH}
""")
''',
'''
def foo():
    return f"""
        {name}
    """
''',
'''
func(arg=f"""
""")
'''
    ],
)
def test_join_formatted_string_columns(input):
    block = PythonBlock(input, auto_flags=True)
    assert block.annotated_ast_node


@pytest.mark.parametrize(
    "input",
    [
        '''b"""
two
""" b"""
four
five
six
"""
''',
        '''
print(b"""
""", sep="")
''',
    ],
)
def test_bytes_concat(input):
    block = PythonBlock(input, auto_flags=True)
    assert block.annotated_ast_node



# pyflyby/_modules.py.
# Copyright (C) 2011, 2012, 2013, 2014, 2015 Karl Chen.
# License: MIT http://opensource.org/licenses/MIT

from __future__ import print_function

import ast
from   functools                import cached_property, total_ordering
import itertools
import os

from   pyflyby._file            import FileText, Filename
from   pyflyby._idents          import DottedIdentifier, is_identifier
from   pyflyby._log             import logger
from   pyflyby._util            import (ExcludeImplicitCwdFromPathCtx, cmp,
                                        memoize, prefixes)

import re
from   six                      import reraise
import sys
import types
from   typing                   import Any, Dict

class ErrorDuringImportError(ImportError):
    """
    Exception raised by import_module if the module exists but an exception
    occurred while attempting to import it.  That nested exception could be
    ImportError, e.g. if a module tries to import another module that doesn't
    exist.
    """


@memoize
def import_module(module_name):
    module_name = str(module_name)
    logger.debug("Importing %r", module_name)
    try:
        result = __import__(module_name, fromlist=['dummy'])
        if result.__name__ != module_name:
            logger.debug("Note: import_module(%r).__name__ == %r",
                         module_name, result.__name__)
        return result
    except ImportError as e:
        # We got an ImportError.  Figure out whether this is due to the module
        # not existing, or whether the module exists but caused an ImportError
        # (perhaps due to trying to import another problematic module).
        # Do this by looking at the exception traceback.  If the previous
        # frame in the traceback is this function (because locals match), then
        # it should be the internal import machinery reporting that the module
        # doesn't exist.  Re-raise the exception as-is.
        # If some sys.meta_path or other import hook isn't compatible with
        # such a check, here are some things we could do:
        #   - Use pkgutil.find_loader() after the fact to check if the module
        #     is supposed to exist.  Note that we shouldn't rely solely on
        #     this before attempting to import, because find_loader() doesn't
        #     work with meta_path.
        #   - Write a memoized global function that compares in the current
        #     environment the difference between attempting to import a
        #     non-existent module vs a problematic module, and returns a
        #     function that uses the working discriminators.
        real_importerror1 = type(e) is ImportError
        real_importerror2 = (sys.exc_info()[2].tb_frame.f_locals is locals())
        m = re.match("^No module named (.*)$", str(e))
        real_importerror3 = (m and m.group(1) == module_name
                             or module_name.endswith("."+m.group(1)))
        logger.debug("import_module(%r): real ImportError: %s %s %s",
                     module_name,
                     real_importerror1, real_importerror2, real_importerror3)
        if real_importerror1 and real_importerror2 and real_importerror3:
            raise
        reraise(ErrorDuringImportError, ErrorDuringImportError(
            "Error while attempting to import %s: %s: %s"
            % (module_name, type(e).__name__, e)), sys.exc_info()[2])
    except Exception as e:
        reraise(ErrorDuringImportError, ErrorDuringImportError(
            "Error while attempting to import %s: %s: %s"
            % (module_name, type(e).__name__, e)), sys.exc_info()[2])


def _my_iter_modules(path, prefix=''):
    # Modified version of pkgutil.ImpImporter.iter_modules(), patched to
    # handle inaccessible subdirectories.
    if path is None:
        return
    try:
        filenames = os.listdir(path)
    except OSError:
        return # silently ignore inaccessible paths
    filenames.sort()  # handle packages before same-named modules
    yielded = {}
    import inspect
    for fn in filenames:
        modname = inspect.getmodulename(fn)
        if modname=='__init__' or modname in yielded:
            continue
        subpath = os.path.join(path, fn)
        ispkg = False
        try:
            if not modname and os.path.isdir(path) and '.' not in fn:
                modname = fn
                for fn in os.listdir(subpath):
                    subname = inspect.getmodulename(fn)
                    if subname=='__init__':
                        ispkg = True
                        break
                else:
                    continue    # not a package
        except OSError:
            continue # silently ignore inaccessible subdirectories
        if modname and '.' not in modname:
            yielded[modname] = 1
            yield prefix + modname, ispkg


def pyc_to_py(filename):
    if filename.endswith(".pyc") or filename.endswith(".pyo"):
        filename = filename[:-1]
    return filename


@total_ordering
class ModuleHandle(object):
    """
    A handle to a module.
    """

    name: DottedIdentifier

    def __new__(cls, arg):
        if isinstance(arg, cls):
            return arg
        if isinstance(arg, Filename):
            return cls._from_filename(arg)
        if isinstance(arg, (str, DottedIdentifier)):
            return cls._from_modulename(arg)
        if isinstance(arg, types.ModuleType):
            return cls._from_module(arg)
        raise TypeError("ModuleHandle: unexpected %s" % (type(arg).__name__,))

    _cls_cache:Dict[Any, Any] = {}

    @classmethod
    def _from_modulename(cls, modulename):
        modulename = DottedIdentifier(modulename)
        try:
            return cls._cls_cache[modulename]
        except KeyError:
            pass
        self = object.__new__(cls)
        self.name = modulename
        cls._cls_cache[modulename] = self
        return self

    @classmethod
    def _from_module(cls, module):
        if not isinstance(module, types.ModuleType):
            raise TypeError
        self = cls._from_modulename(module.__name__)
        assert self.module is module
        return self

    @classmethod
    def _from_filename(cls, filename):
        filename = Filename(filename)
        raise NotImplementedError(
            "TODO: look at sys.path to guess module name")

    @cached_property
    def parent(self):
        if not self.name.parent:
            return None
        return ModuleHandle(self.name.parent)

    @cached_property
    def ancestors(self):
        return tuple(ModuleHandle(m) for m in self.name.prefixes)

    @cached_property
    def module(self):
        """
        Return the module instance.

        :rtype:
          ``types.ModuleType``
        :raise ErrorDuringImportError:
          The module should exist but an error occurred while attempting to
          import it.
        :raise ImportError:
          The module doesn't exist.
        """
        # First check if prefix component is importable.
        if self.parent:
            self.parent.module
        # Import.
        return import_module(self.name)

    @cached_property
    def exists(self):
        """
        Return whether the module exists, according to pkgutil.
        Note that this doesn't work for things that are only known by using
        sys.meta_path.
        """
        name = str(self.name)
        if name in sys.modules:
            return True
        if self.parent and not self.parent.exists:
            return False

        import importlib.util
        find = importlib.util.find_spec

        try:
            pkg = find(name)
        except Exception:
            # Catch all exceptions, not just ImportError.  If the __init__.py
            # for the parent package of the module raises an exception, it'll
            # propagate to here.
            pkg = None
        return pkg is not None

    @cached_property
    def filename(self):
        """
        Return the filename, if appropriate.

        The module itself will not be imported, but if the module is not a
        top-level module/package, accessing this attribute may cause the
        parent package to be imported.

        :rtype:
          `Filename`
        """
        if sys.version_info > (3, 12):
            from importlib.util import find_spec
            try:
                mod = find_spec(str(self.name))
                if mod is None or mod.origin is None:
                    return None
                else:
                    assert isinstance(mod.origin, str)
                    return Filename(mod.origin)
            except ModuleNotFoundError:
                return None
            assert False

        # Use the loader mechanism to find the filename.  We do so instead of
        # using self.module.__file__, because the latter forces importing a
        # module, which may be undesirable.

        import pkgutil
        try:
             #TODO: deprecated and will be removed in 3.14
            loader = pkgutil.get_loader(str(self.name))
        except ImportError:
            return None
        if not loader:
            return None
        # Get the filename using loader.get_filename().  Note that this does
        # more than just loader.filename: for example, it adds /__init__.py
        # for packages.
        if not hasattr(loader, 'get_filename'):
            return None
        filename = loader.get_filename()
        if not filename:
            return None
        return Filename(pyc_to_py(filename))

    @cached_property
    def text(self):
        return FileText(self.filename)

    def __text__(self):
        return self.text

    @cached_property
    def block(self):
        from pyflyby._parse import PythonBlock
        return PythonBlock(self.text)

    @staticmethod
    @memoize
    def list():
        """
        Enumerate all top-level packages/modules.

        :rtype:
          ``tuple`` of `ModuleHandle` s
        """
        import pkgutil
        # Get the list of top-level packages/modules using pkgutil.
        # We exclude "." from sys.path while doing so.  Python includes "." in
        # sys.path by default, but this is undesirable for autoimporting.  If
        # we autoimported random python scripts in the current directory, we
        # could accidentally execute code with side effects.  If the current
        # working directory is /tmp, trying to enumerate modules there also
        # causes problems, because there are typically directories there not
        # readable by the current user.
        with ExcludeImplicitCwdFromPathCtx():
            modlist = pkgutil.iter_modules(None)
            module_names = [t[1] for t in modlist]
        # pkgutil includes all *.py even if the name isn't a legal python
        # module name, e.g. if a directory in $PYTHONPATH has files named
        # "try.py" or "123.py", pkgutil will return entries named "try" or
        # "123".  Filter those out.
        module_names = [m for m in module_names if is_identifier(m)]
        # Canonicalize.
        return tuple(ModuleHandle(m) for m in sorted(set(module_names)))

    @cached_property
    def submodules(self):
        """
        Enumerate the importable submodules of this module.

          >>> ModuleHandle("email").submodules      # doctest:+ELLIPSIS
          (..., ModuleHandle('email.encoders'), ..., ModuleHandle('email.mime'), ...)

        :rtype:
          ``tuple`` of `ModuleHandle` s
        """
        import pkgutil
        module = self.module
        try:
            path = module.__path__
        except AttributeError:
            return ()
        # Enumerate the modules at a given path.  Prefer to use ``pkgutil`` if
        # we can.  However, if it fails due to OSError, use our own version
        # which is robust to that.
        try:
            submodule_names = [t[1] for t in pkgutil.iter_modules(path)]
        except OSError:
            submodule_names = [t[0] for p in path for t in _my_iter_modules(p)]
        return tuple(ModuleHandle("%s.%s" % (self.name,m))
                     for m in sorted(set(submodule_names)))

    @staticmethod
    def _member_from_node(node):
        extractors = {
            # Top-level assignments (as opposed to member assignments
            # whose targets are of type ast.Attribute).
            ast.Assign: lambda x: [t.id for t in x.targets if isinstance(t, ast.Name)],
            ast.ClassDef: lambda x: [x.name],
            ast.FunctionDef: lambda x: [x.name],
        }
        if isinstance(node, tuple(extractors.keys())):
            return extractors[type(node)](node)
        return []

    @cached_property
    def exports(self):
        """
        Get symbols exported by this module.

        Note that this will not recognize symbols that are dynamically
        introduced to the module's namespace or __all__ list.

        :rtype:
          `ImportSet` or ``None``
        :return:
          Exports, or ``None`` if nothing exported.
        """
        from pyflyby._importclns import ImportStatement, ImportSet

        filename = getattr(self, 'filename', None)
        if not filename or not filename.exists:
            # Try to load the module to get the filename
            filename = Filename(self.module.__file__)
        text = FileText(filename)

        ast_mod = ast.parse(str(text), str(filename)).body

        # First, add members that are explicitly defined in the module
        members = list(itertools.chain(*[self._member_from_node(n) \
                                         for n in ast_mod]))

        # If __all__ is defined, try to use it
        all_is_good = False  # pun intended
        all_members = []
        if "__all__" in members:
            # Iterate through the nodes and reconstruct the
            # value of __all__
            for n in ast_mod:
                if isinstance(n, ast.Assign):
                    if "__all__" in self._member_from_node(n):
                        try:
                            all_members = list(ast.literal_eval(n.value))
                            all_is_good = True
                        except (ValueError, TypeError):
                            all_is_good = False
                elif isinstance(n, ast.AugAssign) and \
                     isinstance(n.target, ast.Name) and \
                     n.target.id == "__all__" and all_is_good:
                    try:
                        all_members += list(ast.literal_eval(n.value))
                    except (ValueError, TypeError):
                        all_is_good = False
            if not all(type(s) == str for s in members):
                raise Exception(
                    "Module %r contains non-string entries in __all__"
                    % (str(self.name),))

        if all_is_good:
            members = all_members
        else:
            # Add "from" imports that belong to submodules
            # (note: this will fail to recognize implicit relative imports)
            imp_nodes = [n for n in ast_mod if isinstance(n, ast.ImportFrom)]
            for imp_node in imp_nodes:
                if imp_node.level == 0:
                    from_mod = DottedIdentifier(imp_node.module)
                    if not from_mod.startswith(self.name):
                        continue
                elif imp_node.level == 1 and \
                     filename.base == "__init__.py":
                    # Special case: a relative import can be from a submodule only if
                    # our module's filename is  __init__.py.
                    from_mod = self.name
                    if imp_node.module:
                        from_mod += imp_node.module
                else:
                    continue
                for n in imp_node.names:
                    m  = n.asname or n.name
                    if n.name != "*" and not ModuleHandle(from_mod + m).exists:
                        members.append(m)

        # Filter by non-private.
        members = [n for n in members if not n.startswith("_")]

        # Filter out artificially added "deep" members.
        members = tuple([(n, None) for n in members if "." not in n])
        if not members:
            return None
        return ImportSet(
            [ ImportStatement.from_parts(str(self.name), members) ])

    def __str__(self):
        return str(self.name)

    def __repr__(self):
        return "%s(%r)" % (type(self).__name__, str(self.name))

    def __hash__(self):
        return hash(self.name)

    def __cmp__(self, o):
        if self is o:
            return 0
        if not isinstance(o, ModuleHandle):
            return NotImplemented
        return cmp(self.name, o.name)

    def __eq__(self, o):
        if self is o:
            return True
        if not isinstance(o, ModuleHandle):
            return NotImplemented
        return self.name == o.name

    def __ne__(self, other):
        return not (self == other)

    # The rest are defined by total_ordering
    def __lt__(self, o):
        if not isinstance(o, ModuleHandle):
            return NotImplemented
        return self.name < o.name

    def __getitem__(self, x):
        if isinstance(x, slice):
            return type(self)(self.name[x])
        raise TypeError

    @classmethod
    def containing(cls, identifier):
        """
        Try to find the module that defines a name such as ``a.b.c`` by trying
        to import ``a``, ``a.b``, and ``a.b.c``.

        :return:
          The name of the 'deepest' module (most commonly it would be ``a.b``
          in this example).
        :rtype:
          `Module`
        """
        # In the code below we catch "Exception" rather than just ImportError
        # or AttributeError since importing and __getattr__ing can raise other
        # exceptions.
        identifier = DottedIdentifier(identifier)
        try:
            module = ModuleHandle(identifier[:1])
            result = module.module
        except Exception as e:
            raise ImportError(e)
        # TODO: as far as I can tell the code here is never reached, or haven't
        # been in quite some time as the line below was invalid on Python 3 since 2011
        # zip(...)[...] fails as zip is not indexable.
        # the only place that seem to be using this method is XrefScanner.
        for part, prefix in list(zip(identifier, prefixes(identifier)))[1:]:
            try:
                result = getattr(result, str(part))
            except Exception:
                try:
                    module = cls(prefix)
                    result = module.module
                except Exception as e:
                    raise ImportError(e)
            else:
                if isinstance(result, types.ModuleType):
                    module = cls(result)
        logger.debug("Imported %r to get %r", module, identifier)
        return module



"""
This module contain utility functions to sort imports in a Python Block.
"""
from __future__ import annotations

from collections import Counter
from dataclasses import dataclass
from itertools import groupby
from pyflyby._importstmt import ImportStatement, PythonStatement
from pyflyby._parse import PythonBlock
from typing import List, Tuple, Union


# @dataclass
# class ImportSection:
#     """
#     This represent an Import
#     """
#     sections: List[ImportGroup]


@dataclass
class ImportGroup:
    """
    Typically at the top of a file, the first import will be part
    of an ImportGroup, and subsequent imports
    will be other import groups which.

    Import sorting will affect only the imports in the same group,
    as we want import sorting to be indempotent. If import sorting
    were migrating imports between groups, then imports could be
    moved to sections were comments would nto be relevant.
    """

    imports: List[ImportStatement]

    @classmethod
    def from_statements(cls, statements: List[PythonStatement]) -> ImportGroup:
        return ImportGroup([ImportStatement(s) for s in statements])

    def sorted(self) -> ImportGroup:
        """
        return an ImportGroup with import sorted lexicographically.
        """
        return ImportGroup(sorted(self.imports, key=lambda x: x._cmp()))

    def sorted_subgroups(self) -> List[Tuple[bool, List[ImportStatement]]]:
        """
        Return a list of subgroup keyed by module and whether they are the sole import
        from this module.

        From issue #13, we will want to sort import from the same package together when
        there is more than one import and separat with a blank line

        We also group all imports from a package that appear only once together.

        For this we need both to know if the import is from a single import (the boolean).

        Returns
        -------
        bool:
           wether from a single import
        List[ImportStatement]
           The actual import.
        """
        c = Counter(imp.module[0] for imp in self.imports)
        return [
            (c[k] > 1, list(v)) for k, v in groupby(self.imports, lambda x: x.module[0])
        ]


def split_import_groups(
    statements: Tuple[PythonStatement],
) -> List[Union[ImportGroup, PythonStatement]]:
    """
    Given a list of statements split into import groups.

    One of the question is how to treat split with comments.

     - Do blank lines after groups with comments start a new block
     - Does the comment line create a complete new block.

    In particular because we want this to be indempotent,
    we can't move imports between groups.
    """

    # these are the import groups we'll
    groups: List[PythonStatement | ImportGroup] = []

    current_group: List[PythonStatement] = []
    statemt_iterator = iter(statements)
    for statement in statemt_iterator:
        if statement.is_blank and current_group:
            pass
            # currently do nothing with whitespace while in import groups.
        elif statement.is_import:
            # push on top of current_comment.
            current_group.append(statement)
        else:
            if current_group:
                groups.append(ImportGroup.from_statements(current_group))
            current_group = []
            groups.append(statement)

            # we should break of and populate rest
            # We can't do anything if we encounter any non-import statement,
            # as we do no know if it can be a conditional.
            # technically I guess we coudl find another import block, and reorder these.
    if current_group:
        groups.append(ImportGroup.from_statements(current_group))

    # this is an iterator, not an iterable, we exaust it to reify the rest.

    # first group may be empty if the first line is a comment.
    # We filter and sort relevant statements.
    groups = [g for g in groups if groups]
    sorted_groups: List[PythonStatement | ImportGroup] = [
        g.sorted() if isinstance(g, ImportGroup) else g for g in groups
    ]
    return sorted_groups


def regroup(groups: List[ImportGroup | PythonStatement]) -> PythonBlock:
    """
    given import groups and list of statement, return an Python block with sorted import
    """
    res: str = ""
    in_single = False
    for group in groups:
        if isinstance(group, ImportGroup):
            # the subgroup here will be responsible for  groups reordering.
            for mult, subgroup in group.sorted_subgroups():
                if mult:
                    if in_single:
                        res += "\n"
                    if not res.endswith("\n"):
                        res += "\n"
                    for x in subgroup:
                        res += x.pretty_print(import_column=30).rstrip() + "\n"
                    if not res.endswith("\n\n"):
                        res += "\n"
                    in_single = False
                else:
                    assert len(subgroup) == 1
                    in_single = True
                    for sub in subgroup:
                        res += str(sub).rstrip() + "\n"

            if not res.endswith("\n\n"):
                res += "\n"
        else:
            if in_single and not res.endswith("\n\n"):
                res += "\n"
            in_single = False
            res += str(PythonBlock(group))

    return PythonBlock.concatenate([PythonBlock(res)])


def sort_imports(block: PythonBlock) -> PythonBlock:
    assert isinstance(block, PythonBlock)
    # we ignore below that block.statement can be a List[Unknown]
    gs = split_import_groups(block.statements)  # type: ignore
    # TODO: math the other fileds like filename....
    return regroup(gs)



# pyflyby/test_importstmt.py

# License for THIS FILE ONLY: CC0 Public Domain Dedication
# http://creativecommons.org/publicdomain/zero/1.0/


from   pytest                   import raises
from   unittest.mock            import patch

from   pyflyby._flags           import CompilerFlags
from   pyflyby._format          import FormatParams
from   pyflyby._importstmt      import (Import, ImportSplit, ImportStatement,
                                        read_black_config)

def test_Import_from_parts_1():
    imp = Import.from_parts(".foo.bar", "bar")
    assert imp.fullname  == ".foo.bar"
    assert imp.import_as == "bar"
    assert imp.split     == ImportSplit(".foo", "bar", None)
    assert str(imp)      == "from .foo import bar"


def test_Import_from_split_1():
    imp = Import(ImportSplit(".foo", "bar", None))
    assert imp.fullname  == ".foo.bar"
    assert imp.import_as == "bar"
    assert imp.split     == ImportSplit(".foo", "bar", None)
    assert str(imp)      == "from .foo import bar"


def test_Import_from_Statement_1():
    imp = Import(ImportStatement("from foo import bar"))
    assert imp.fullname  == "foo.bar"
    assert imp.import_as == "bar"
    assert imp.split     == ImportSplit("foo", "bar", None)
    assert str(imp)      == "from foo import bar"
    assert imp == Import("from foo import bar")


def test_Import_basic_1():
    imp = Import("from foo.foof import bar")
    assert imp.fullname  == "foo.foof.bar"
    assert imp.import_as == "bar"
    assert imp.split     == ImportSplit("foo.foof", "bar", None)
    assert str(imp)      == "from foo.foof import bar"


def test_Import_relative_1():
    imp = Import("from .foo import bar")
    assert imp.fullname  == ".foo.bar"
    assert imp.import_as == "bar"
    assert imp.split     == ImportSplit(".foo", "bar", None)
    assert str(imp)      == "from .foo import bar"


def test_Import_relative_local_1():
    imp = Import("from . import foo")
    assert imp.fullname  == ".foo"
    assert imp.import_as == "foo"
    assert imp.split     == ImportSplit(".", "foo", None)
    assert str(imp)      == "from . import foo"


def test_Import_module_1():
    imp = Import("import   foo . bar")
    assert imp.fullname  == "foo.bar"
    assert imp.import_as == "foo.bar"
    assert imp.split     == ImportSplit(None, "foo.bar", None)
    assert str(imp)      == "import foo.bar"


def test_Import_import_as_1():
    imp = Import("import   foo . bar  as  baz")
    assert imp.fullname  == "foo.bar"
    assert imp.import_as == "baz"
    assert imp.split     == ImportSplit("foo", "bar", "baz")
    assert str(imp)      == "from foo import bar as baz"


def test_Import_import_as_same_1():
    imp = Import("import   foo . bar  as  bar")
    assert imp.fullname  == "foo.bar"
    assert imp.import_as == "bar"
    assert imp.split     == ImportSplit("foo", "bar", None)
    assert str(imp)      == "from foo import bar"
    assert imp == Import("from foo import bar")


def test_Import_eqne_1():
    imp1a = Import("from foo import bar")
    imp1b = Import("from foo import bar")
    imp2  = Import("from .foo import bar")
    assert     (imp1a == imp1b)
    assert not (imp1a != imp1b)
    assert     (imp1a != imp2 )
    assert not (imp1a == imp2 )


def test_Import_eqne_2():
    imp1a = Import("from foo import bar")
    imp1b = Import("from foo import bar")
    imp2  = Import("from foo import bar as Bar")
    assert     (imp1a == imp1b)
    assert not (imp1a != imp1b)
    assert     (imp1a != imp2 )
    assert not (imp1a == imp2 )


def test_Import_prefix_match_1():
    result = Import("import ab.cd.ef").prefix_match(Import("import ab.cd.xy"))
    assert result == ('ab', 'cd')


def test_Import_replace_1():
    result = Import("from aa.bb import cc").replace("aa.bb", "xx.yy")
    assert result == Import('from xx.yy import cc')


def test_Import_replace_2():
    result = Import("from aa import bb").replace("aa.bb", "xx.yy")
    assert result == Import('from xx import yy as bb')


@patch("pyflyby._importstmt.read_black_config", lambda: {"line_length": 20})
def test_Import_black_line_length():
    stmt = Import("from a123456789 import b123456789")
    result = stmt.pretty_print(params=FormatParams(use_black=True))
    assert result == "from a123456789 import (\n    b123456789,\n)\n"


def test_ImportStatement_1():
    stmt = ImportStatement("import  foo . bar")
    assert stmt.fromname == None
    assert stmt.aliases == (("foo.bar", None),)
    assert stmt.imports == (Import(ImportSplit(None, "foo.bar", None)),)
    assert str(stmt) == "import foo.bar"


def test_ImportStatement_member_1():
    stmt = ImportStatement("from  foo  import  bar ")
    assert stmt.fromname == "foo"
    assert stmt.aliases == (("bar", None),)
    assert stmt.imports == (Import(ImportSplit("foo", "bar", None)),)
    assert str(stmt) == "from foo import bar"


def test_ImportStatement_multi_1():
    stmt = ImportStatement("from foo  import bar, bar2, bar")
    assert stmt.fromname == "foo"
    assert stmt.aliases == (("bar", None), ("bar2", None), ("bar", None))
    assert stmt.imports == (Import(ImportSplit("foo", "bar", None)),
                            Import(ImportSplit("foo", "bar2", None)),
                            Import(ImportSplit("foo", "bar", None)))
    assert str(stmt) == "from foo import bar, bar2, bar"


def test_ImportStatement_alias_1():
    stmt = ImportStatement("from foo  import bar  as  bar,   bar as   baz")
    assert stmt.fromname == "foo"
    assert stmt.aliases == (("bar", "bar"), ("bar", "baz"))
    assert stmt.imports == (Import(ImportSplit("foo", "bar", "bar")),
                            Import(ImportSplit("foo", "bar", "baz")))
    assert str(stmt) == "from foo import bar as bar, bar as baz"


def test_ImportStatement_deep_member_1():
    stmt = ImportStatement("from foo.bar import baz")
    assert stmt.fromname == "foo.bar"
    assert stmt.aliases == (("baz", None),)
    assert stmt.imports == (Import(ImportSplit("foo.bar", "baz", None)),)
    assert str(stmt) == "from foo.bar import baz"


def test_ImportStatement_relative_1():
    stmt = ImportStatement("from .foo import bar")
    assert stmt.fromname == ".foo"
    assert stmt.aliases == (("bar", None),)
    assert stmt.imports == (Import(ImportSplit(".foo", "bar", None)),)
    assert str(stmt) == "from .foo import bar"


def test_ImportStatement_relative_local_1():
    stmt = ImportStatement("from .  import bar , bar2 as baz2")
    assert stmt.fromname == "."
    assert stmt.aliases == (("bar", None), ("bar2", "baz2"))
    assert stmt.imports == (Import(ImportSplit(".", "bar", None)),
                            Import(ImportSplit(".", "bar2", "baz2")))
    assert str(stmt) == "from . import bar, bar2 as baz2"


def test_ImportStatement_flags_1():
    stmt = ImportStatement("from __future__ import division, print_function")
    assert stmt.flags == CompilerFlags('division', 'print_function')


def test_ImportStatement_flags_2():
    stmt = ImportStatement("from _future__ import division, print_function")
    assert stmt.flags == CompilerFlags.from_int(0)


def test_ImportStatement_eqne_1():
    stmt1a = ImportStatement("from a import b"   )
    stmt1b = ImportStatement("from a import b"   )
    stmt2  = ImportStatement("from a import b, b")
    assert     (stmt1a == stmt1b)
    assert not (stmt1a != stmt1b)
    assert     (stmt1a != stmt2 )
    assert not (stmt1a == stmt2 )


def test_ImportStatement_eqne_2():
    stmt1a = ImportStatement("from a import b"   )
    stmt1b = ImportStatement("from a import b"   )
    stmt2  = ImportStatement("from a import b as b")
    assert     (stmt1a == stmt1b)
    assert not (stmt1a != stmt1b)
    assert     (stmt1a != stmt2 )
    assert not (stmt1a == stmt2 )


@patch("black.files.find_pyproject_toml", lambda root: None)
def test_ImportStatement_pretty_print_black_no_config():
    # running should not error out when no pyproject.toml file is found
    stmt = ImportStatement("from a import b")
    result = stmt.pretty_print(params=FormatParams(use_black=True))
    assert isinstance(result, str)


@patch("black.files.find_pyproject_toml", lambda root: None)
def test_read_black_config_no_config():
    # reading black config should work when no pyproject.toml file is found
    config = read_black_config()
    assert config == {}


@patch("black.files.find_pyproject_toml", lambda root: "pyproject.toml")
@patch(
    "black.files.parse_pyproject_toml",
    lambda path: {
        "line_length": 80,
        "skip_magic_trailing_comma": True,
        "skip_string_normalization": False,
        "skip_source_first_line": True
    }
)
def test_read_black_config_extracts_config_subset():
    config = read_black_config()
    # should copy the desired black options
    assert config["line_length"] == 80
    assert config["skip_magic_trailing_comma"] == True
    assert config["skip_string_normalization"] == False
    # should not copy anything else
    assert "skip_source_first_line" not in config


@patch("black.files.find_pyproject_toml", lambda root: "pyproject.toml")
@patch("black.files.parse_pyproject_toml", lambda path: {"target_version": ["py310", "py311"]})
def test_read_black_config_target_version_list():
    config = read_black_config()
    assert config["target_version"] == {"py310", "py311"}


@patch("black.files.find_pyproject_toml", lambda root: "pyproject.toml")
@patch("black.files.parse_pyproject_toml", lambda path: {"target_version": "py311"})
def test_read_black_config_target_version_str():
    config = read_black_config()
    assert config["target_version"] == "py311"

@patch("black.files.find_pyproject_toml", lambda root: "pyproject.toml")
@patch("black.files.parse_pyproject_toml", lambda path: {"target_version": object()})
def test_read_black_config_target_version_other():
    with raises(ValueError, match="Invalid config for black"):
        read_black_config()
